from typing import Any, Callable, MutableMapping, Optional

from tensorflow import keras


class HyperparameterScheduler(keras.callbacks.Callback):
    """Generic hyperparameter scheduler.

    !!! example
        ```python
        bop = lq.optimizers.Bop(threshold=1e-6, gamma=1e-3)
        adam = tf.keras.optimizers.Adam(0.01)
        optimizer = lq.optimizers.CaseOptimizer(
            (lq.optimizers.Bop.is_binary_variable, bop), default_optimizer=adam,
        )
        callbacks = [
            HyperparameterScheduler(lambda x: 0.001 * (0.1 ** (x // 30)), "gamma", bop)
        ]
        ```

    # Arguments
        schedule: a function that takes an epoch index as input
            (integer, indexed from 0) and returns a new hyperparameter as output.
        hyperparameter: str. the name of the hyperparameter to be scheduled.
        optimizer: the optimizer that contains the hyperparameter that will be scheduled.
            Defaults to `self.model.optimizer` if `optimizer == None`.
        update_freq: str (optional), denotes on what update_freq to change the
            hyperparameter. Can be either "epoch" (default) or "step".
        verbose: int. 0: quiet, 1: update messages.
        log_name: str (optional), under which name to log this hyperparameter to
            Tensorboard. If `None`, defaults to `hyperparameter`. Use this if you have
            several schedules for the same hyperparameter on different optimizers.
    """

    def __init__(
        self,
        schedule: Callable,
        hyperparameter: str,
        optimizer: Optional[keras.optimizers.Optimizer] = None,
        update_freq: str = "epoch",
        verbose: int = 0,
        log_name: Optional[str] = None,
    ):
        super(HyperparameterScheduler, self).__init__()
        self.optimizer = optimizer
        self.schedule = schedule
        self.hyperparameter = hyperparameter
        self.log_name = log_name or hyperparameter
        self.verbose = verbose

        if update_freq not in ["epoch", "step"]:
            raise ValueError(
                "HyperparameterScheduler.update_freq can only be 'step' or 'epoch'."
                f" Received value '{update_freq}'"
            )

        self.update_freq = update_freq

    def set_model(self, model: keras.models.Model) -> None:
        super().set_model(model)
        if self.optimizer is None:
            # It is not possible for a model to reach this state and not have
            # an optimizer, so we can safely access it here.
            self.optimizer = model.optimizer

        if not hasattr(self.optimizer, self.hyperparameter):
            raise ValueError(
                f'Optimizer must have a "{self.hyperparameter}" attribute.'
            )

    def set_hyperparameter(self, t: int) -> Any:
        hp = getattr(self.optimizer, self.hyperparameter)
        try:  # new API
            hyperparameter_val = keras.backend.get_value(hp)
            hyperparameter_val = self.schedule(t, hyperparameter_val)
        except TypeError:  # Support for old API for backward compatibility
            hyperparameter_val = self.schedule(t)
        keras.backend.set_value(hp, hyperparameter_val)
        return hp

    def on_batch_begin(
        self, batch: int, logs: Optional[MutableMapping[str, Any]] = None
    ) -> None:
        if not self.update_freq == "step":
            return

        # We use optimizer.iterations (i.e. global step), since batch only
        # reflects the batch index in the current epoch.
        batch = keras.backend.get_value(self.optimizer.iterations)
        hp = self.set_hyperparameter(batch)

        if self.verbose > 0:
            print(
                f"Batch {batch}: {self.log_name} is now {keras.backend.get_value(hp)}."
            )

    def on_epoch_begin(
        self, epoch: int, logs: Optional[MutableMapping[str, Any]] = None
    ) -> None:
        if not self.update_freq == "epoch":
            return

        hp = self.set_hyperparameter(epoch)

        if self.verbose > 0:
            print(
                f"Epoch {epoch}: {self.log_name} is now {keras.backend.get_value(hp)}."
            )

    def on_epoch_end(
        self, epoch: int, logs: Optional[MutableMapping[str, Any]] = None
    ) -> None:
        logs = logs or {}
        hp = getattr(self.optimizer, self.hyperparameter)
        logs[self.log_name] = keras.backend.get_value(hp)



"""Context managers that configure global behaviour of Larq."""

import contextlib
import threading

__all__ = [
    "metrics_scope",
    "quantized_scope",
    "get_training_metrics",
    "should_quantize",
]


_quantized_scope = threading.local()
_quantized_scope.should_quantize = False


@contextlib.contextmanager
def quantized_scope(quantize):
    """A context manager to define the behaviour of `QuantizedVariable`.

    !!! example
        ```python
        model.save("full_precision_model.h5")  # save full precision latent weights
        fp_weights = model.get_weights()  # get latent weights

        with larq.context.quantized_scope(True):
            model.save("binary_model.h5")  # save binarized weights
            weights = model.get_weights()  # get binarized weights
        ```

    # Arguments
        quantize: If `should_quantize` is `True`, `QuantizedVariable` will return their
            quantized value in the forward pass. If `False`, `QuantizedVariable` will
            act as a latent variable.
    """
    backup = should_quantize()
    _quantized_scope.should_quantize = quantize
    yield quantize
    _quantized_scope.should_quantize = backup


def should_quantize():
    """Returns the current quantized scope."""
    return getattr(_quantized_scope, "should_quantize", False)


_global_training_metrics = set()
_available_metrics = {"flip_ratio"}


@contextlib.contextmanager
def metrics_scope(metrics=[]):
    """A context manager to set the training metrics to be used in quantizers.

    !!! example
        ```python
        with larq.context.metrics_scope(["flip_ratio"]):
            model = tf.keras.models.Sequential(
                [larq.layers.QuantDense(3, kernel_quantizer="ste_sign", input_shape=(32,))]
            )
        model.compile(loss="mse", optimizer="sgd")
        ```

    # Arguments
        metrics: Iterable of metrics to add to quantizers defined inside this context.
            Currently only the `flip_ratio` metric is available.
    """
    for metric in metrics:
        if metric not in _available_metrics:
            raise ValueError(
                f"Unknown training metric '{metric}'. Available metrics: {_available_metrics}."
            )
    backup = _global_training_metrics.copy()
    _global_training_metrics.update(metrics)
    yield _global_training_metrics
    _global_training_metrics.clear()
    _global_training_metrics.update(backup)


def get_training_metrics():
    """Retrieves a live reference to the training metrics in the current scope.

    Updating and clearing training metrics using `larq.context.metrics_scope` is
    preferred, but `get_training_metrics` can be used to directly access them.

    !!! example
        ```python
        get_training_metrics().clear()
        get_training_metrics().add("flip_ratio")
        ```

    # Returns
        A set of training metrics in the current scope.
    """
    return _global_training_metrics



import pytest
import tensorflow as tf

from larq import context


def test_eager_and_graph_mode_fixture(eager_and_graph_mode):
    if eager_and_graph_mode == "eager":
        assert tf.executing_eagerly()
    else:
        assert not tf.executing_eagerly()
        assert tf.compat.v1.get_default_session() is not None


@pytest.mark.usefixtures("eager_mode")
def test_eager_mode_fixture():
    assert tf.executing_eagerly()


@pytest.mark.usefixtures("graph_mode")
def test_graph_mode_fixture():
    assert not tf.executing_eagerly()
    assert tf.compat.v1.get_default_session() is not None


def test_distribute_scope(distribute_scope):
    assert tf.distribute.has_strategy() is distribute_scope


def test_quantize_scope(quantized):
    assert context.should_quantize() == quantized



import itertools
from dataclasses import dataclass
from typing import Any, Callable, Iterator, Mapping, Optional, Sequence, TypeVar, Union

import numpy as np
import tensorflow as tf
from terminaltables import AsciiTable

from larq import layers as lq_layers
from larq.utils import memory_as_readable_str

__all__ = ["summary"]

mac_containing_layers = (
    lq_layers.QuantConv2D,
    lq_layers.QuantSeparableConv2D,
    lq_layers.QuantDepthwiseConv2D,
    lq_layers.QuantDense,
    tf.keras.layers.Conv2D,
    tf.keras.layers.SeparableConv2D,
    tf.keras.layers.DepthwiseConv2D,
    tf.keras.layers.Dense,
    lq_layers.QuantConv1D,
    lq_layers.QuantSeparableConv1D,
    tf.keras.layers.Conv1D,
    tf.keras.layers.SeparableConv1D,
)

op_count_supported_layer_types = (
    tf.keras.layers.Flatten,
    tf.keras.layers.BatchNormalization,
    tf.keras.layers.MaxPool2D,
    tf.keras.layers.AveragePooling2D,
    tf.keras.layers.MaxPool1D,
    tf.keras.layers.AveragePooling1D,
    *mac_containing_layers,
)

T = TypeVar("T")


def _flatten(lst: Iterator[Iterator[T]]) -> Sequence[T]:
    return list(itertools.chain.from_iterable(lst))


def _bitsize_as_str(bitsize: int) -> str:
    bitsize_names = {8: "byte", 8 * 1024: "kB"}

    try:
        return bitsize_names[bitsize]
    except KeyError:
        raise NotImplementedError()


def _number_as_readable_str(num: float) -> str:
    # The initial rounding here is necessary so that e.g. `999000` gets
    # formatted as `1.000 M` rather than `1000 k`
    num = float(f"{num:.3g}")

    # For numbers less than 1000, output them directly, stripping any trailing
    # zeros and decimal places.
    if num < 1000:
        return str(num).rstrip("0").rstrip(".")

    # For numbers that are at least 1000 trillion (1 quadrillion) format with
    # scientific notation (3 s.f. = 2 d.p. in scientific notation).
    if num >= 1e15:
        return f"{num:.2E}"

    # Count the magnitude.
    magnitude = 0
    while abs(num) >= 1000 and magnitude < 4:
        magnitude += 1
        num /= 1000.0

    # ':.3g' formats the number with 3 significant figures, without stripping trailing
    # zeros.
    num = f"{num:.3g}".rstrip(".")
    unit = ["", " k", " M", " B", " T"][magnitude]
    return num + unit


def _format_table_entry(x: float, units: int = 1) -> Union[float, str]:
    try:
        assert not np.isnan(x)
        if type(x) == str or x == 0 or units == 1:
            return x
        return x / units
    except Exception:
        return "?"


def _normalize_shape(shape):
    return tuple(dim if dim else -1 for dim in shape)


class WeightProfile:
    def __init__(self, weight, trainable: bool = True):
        self._weight = weight
        self.bitwidth = getattr(weight, "precision", 32)
        self.trainable = trainable

    @property
    def count(self) -> int:
        return int(np.prod(self._weight.shape.as_list()))

    @property
    def memory(self) -> int:
        return self.bitwidth * self.count

    @property
    def fp_equivalent_memory(self) -> int:
        return 32 * self.count

    @property
    def int8_fp_weights_memory(self) -> int:
        """Count any 32- or 16-bit weights as 8 bits instead."""

        if self.bitwidth > 8:
            return self.count * 8
        return self.bitwidth * self.count

    def is_bias(self) -> bool:
        return "bias" in self._weight.name


@dataclass
class OperationProfile:
    n: int
    precision: int
    op_type: str


class LayerProfile:
    def __init__(self, layer: tf.keras.layers.Layer):
        self._layer = layer

        weights = layer.weights
        if isinstance(layer, tf.keras.layers.BatchNormalization):
            fused_pairs = [("beta", "moving_mean"), ("gamma", "moving_variance")]
            for pair in fused_pairs:
                names = [w.name.split("/")[-1].replace(":0", "") for w in weights]
                if pair[0] in names and pair[1] in names:
                    weights.pop(names.index(pair[0]))

        self.weight_profiles = [
            WeightProfile(
                weight,
                trainable=any(weight is w for w in layer.trainable_weights),
            )
            for weight in weights
        ]

        self.op_profiles = []

        if isinstance(layer, mac_containing_layers) and self.output_pixels:
            for p in self.weight_profiles:
                if not p.is_bias():
                    self.op_profiles.append(
                        OperationProfile(
                            n=p.count * self.output_pixels,
                            precision=max(self.input_precision or 32, p.bitwidth),
                            op_type="mac",
                        )
                    )

    @property
    def memory(self) -> int:
        return sum(p.memory for p in self.weight_profiles)

    @property
    def int8_fp_weights_memory(self) -> int:
        return sum(p.int8_fp_weights_memory for p in self.weight_profiles)

    @property
    def fp_equivalent_memory(self) -> int:
        return sum(p.fp_equivalent_memory for p in self.weight_profiles)

    def weight_count(
        self, bitwidth: Optional[int] = None, trainable: Optional[bool] = None
    ) -> int:
        count = 0
        for p in self.weight_profiles:
            if (bitwidth is None or p.bitwidth == bitwidth) and (
                trainable is None or p.trainable == trainable
            ):
                count += p.count
        return count

    def op_count(
        self, op_type: Optional[str] = None, precision: Optional[int] = None
    ) -> Optional[int]:
        if op_type != "mac":
            raise ValueError("Currently only counting of MAC-operations is supported.")

        if (
            isinstance(self._layer, op_count_supported_layer_types)
            and self.output_pixels
        ):
            count = 0
            for op in self.op_profiles:
                if (precision is None or op.precision == precision) and (
                    op_type is None or op.op_type == op_type
                ):
                    count += op.n
            return count
        return None

    @property
    def input_precision(self) -> Optional[int]:
        try:
            return self._layer.input_quantizer.precision
        except AttributeError:
            return None

    @property
    def output_shape(self) -> Optional[Sequence[int]]:
        try:
            output_shape = self._layer.output_shape
            if isinstance(output_shape, list):
                if len(output_shape) == 1:
                    return _normalize_shape(output_shape[0])
                return [_normalize_shape(shape) for shape in output_shape]
            return _normalize_shape(output_shape)
        except AttributeError:
            return None

    @property
    def output_shape_str(self) -> str:
        try:
            return str(self.output_shape or "multiple")
        except RuntimeError:
            return "?"

    @property
    def output_pixels(self) -> Optional[int]:
        """Number of pixels for a single feature map (1 for fully connected layers)."""
        if not self.output_shape:
            return None
        if len(self.output_shape) == 4:
            return int(np.prod(self.output_shape[1:3]))
        if len(self.output_shape) == 3:
            return self.output_shape[1]
        if len(self.output_shape) == 2:
            return 1
        raise NotImplementedError()

    @property
    def unique_param_bidtwidths(self) -> Sequence[int]:
        return sorted(set([p.bitwidth for p in self.weight_profiles]))

    @property
    def unique_op_precisions(self) -> Sequence[int]:
        return sorted(set([op.precision for op in self.op_profiles]))

    def generate_table_row(
        self, table_config: Mapping[str, Any]
    ) -> Sequence[Union[str, float]]:
        row = [self._layer.name, self.input_precision or "-", self.output_shape_str]
        for i in table_config["param_bidtwidths"]:
            n = self.weight_count(i)
            n = _format_table_entry(n, table_config["param_units"])
            row.append(n)
        row.append(_format_table_entry(self.memory, table_config["memory_units"]))
        for i in table_config["mac_precisions"]:
            n = self.op_count("mac", i)
            n = _format_table_entry(n, table_config["mac_units"])
            row.append(n)
        return row


class ModelProfile(LayerProfile):
    def __init__(self, model: tf.keras.models.Model):
        self.layer_profiles = [LayerProfile(layer) for layer in model.layers]

    @property
    def memory(self) -> int:
        return sum(lp.memory for lp in self.layer_profiles)

    @property
    def int8_fp_weights_memory(self) -> int:
        return sum(lp.int8_fp_weights_memory for lp in self.layer_profiles)

    @property
    def fp_equivalent_memory(self) -> int:
        return sum(lp.fp_equivalent_memory for lp in self.layer_profiles)

    def weight_count(
        self, bitwidth: Optional[int] = None, trainable: Optional[bool] = None
    ) -> int:
        return sum(lp.weight_count(bitwidth, trainable) for lp in self.layer_profiles)

    def op_count(
        self, op_type: Optional[str] = None, bitwidth: Optional[int] = None
    ) -> int:
        return sum(lp.op_count(op_type, bitwidth) or 0 for lp in self.layer_profiles)

    @property
    def unique_param_bidtwidths(self) -> Sequence[int]:
        return sorted(
            set(_flatten(lp.unique_param_bidtwidths for lp in self.layer_profiles))
        )

    @property
    def unique_op_precisions(self) -> Sequence[int]:
        return sorted(
            set(_flatten(lp.unique_op_precisions for lp in self.layer_profiles))
        )

    def _generate_table_header(self, table_config: Mapping[str, Any]) -> Sequence[str]:
        return [
            "Layer",
            "Input prec.\n(bit)",
            "Outputs",
            *(
                f"# {i}-bit\nx {table_config['param_units']}"
                for i in table_config["param_bidtwidths"]
            ),
            f"Memory\n({_bitsize_as_str(table_config['memory_units'])})",
            *(f"{i}-bit MACs" for i in table_config["mac_precisions"]),
        ]

    def _generate_table_total(
        self, table_config: Mapping[str, Any]
    ) -> Sequence[Union[float, str]]:
        row = ["Total", "", ""]
        for i in table_config["param_bidtwidths"]:
            row.append(
                _format_table_entry(self.weight_count(i), table_config["param_units"])
            )
        row.append(_format_table_entry(self.memory, table_config["memory_units"]))
        for i in table_config["mac_precisions"]:
            row.append(
                _format_table_entry(self.op_count("mac", i), table_config["mac_units"])
            )
        return row

    def generate_table(
        self, include_macs: bool = True
    ) -> Sequence[Sequence[Union[float, str]]]:
        table_config = {
            "param_bidtwidths": self.unique_param_bidtwidths,
            "mac_precisions": self.unique_op_precisions if include_macs else [],
            "param_units": 1,
            "memory_units": 8 * 1024,
            "mac_units": 1,
        }

        table = []

        table.append(self._generate_table_header(table_config))

        for lp in self.layer_profiles:
            table.append(lp.generate_table_row(table_config))

        table.append(self._generate_table_total(table_config))

        return table

    def generate_summary(
        self, include_macs: bool = True
    ) -> Sequence[Sequence[Union[str, float]]]:
        summary = [
            ["Total params", _number_as_readable_str(self.weight_count())],
            [
                "Trainable params",
                _number_as_readable_str(self.weight_count(trainable=True)),
            ],
            [
                "Non-trainable params",
                _number_as_readable_str(self.weight_count(trainable=False)),
            ],
            ["Model size", memory_as_readable_str(self.memory)],
            [
                "Model size (8-bit FP weights)",
                memory_as_readable_str(self.int8_fp_weights_memory),
            ],
            ["Float-32 Equivalent", memory_as_readable_str(self.fp_equivalent_memory)],
            [
                "Compression Ratio of Memory",
                self.memory / max(1e-8, self.fp_equivalent_memory),
            ],
        ]

        if include_macs:
            binarization_ratio = self.op_count("mac", 1) / max(
                1, self.op_count(op_type="mac")
            )
            ternarization_ratio = self.op_count("mac", 2) / max(
                1, self.op_count(op_type="mac")
            )
            summary.append(
                [
                    "Number of MACs",
                    _number_as_readable_str(self.op_count(op_type="mac")),
                ]
            )
            if binarization_ratio > 0:
                summary.append(
                    ["Ratio of MACs that are binarized", f"{binarization_ratio:.4f}"]
                )
            if ternarization_ratio > 0:
                summary.append(
                    ["Ratio of MACs that are ternarized", f"{ternarization_ratio:.4f}"]
                )

        return summary


def sanitize_table(table_data: Sequence[Sequence[Any]]) -> Sequence[Sequence[str]]:
    return [
        [f"{v:.2f}" if type(v) == float else str(v) for v in row] for row in table_data
    ]


class LayersTable(AsciiTable):
    def __init__(self, table_data, title=None):
        super().__init__(sanitize_table(table_data), title=title)
        self.inner_column_border = False
        self.justify_columns = {
            i: "left" if i == 0 else "right" for i in range(len(table_data[0]))
        }
        self.inner_footing_row_border = True
        self.inner_heading_row_border = True


class SummaryTable(AsciiTable):
    def __init__(self, table_data, title=None):
        super().__init__(sanitize_table(table_data), title=title)
        self.inner_column_border = False
        self.inner_heading_row_border = False


def summary(
    model: tf.keras.models.Model,
    print_fn: Callable[[str], Any] = None,
    include_macs: bool = True,
) -> None:
    """Prints a string summary of the network.

    The summary includes the following information per layer:

    - input precision,
    - output dimension,
    - weight count (broken down by bidtwidth),
    - memory footprint in kilobytes (`8*1024` 1-bit weights = 1 kB),
    - number of multiply-accumulate (MAC) operations broken down by precision (*optional & expermental*).

    A single MAC operation contains both a multiplication and an addition. The precision
    of a MAC operation is defined as the maximum bitwidth of its inputs.

    Additionally, the following overall statistics for the model are supplied:

    - total number of weights,
    - total number of trainable weights,
    - total number of non-trainable weights,
    - model size,
    - model size (8-bit FP weights): memory footprint if FP weights were 8 bit,
    - float-32 equivalent size: memory footprint if all weights were 32 bit,
    - compression ratio achieved by quantizing weights,
    - total number of MAC operations,
    - ratio of MAC operations that is binarized and can be accelated with XNOR-gates.

    # Arguments
        model: model instance.
        print_fn: Print function to use. Defaults to `print`. You can set it to a custom
            function in order to capture the string summary.
        include_macs: whether or not to include the number of MAC-operations in the
            summary.

    # Raises
        ValueError: if called before the model is built.
    """

    if not model.built:
        raise ValueError(
            "This model has not yet been built. Build the model first by calling "
            "`model.build()` or calling `model.fit()` with some data, or specify an "
            "`input_shape` argument in the first layer(s) for automatic build."
        )

    if not print_fn:
        print_fn = print

    model_profile = ModelProfile(model)
    print_fn(
        LayersTable(model_profile.generate_table(), title=f"{model.name} stats").table
    )
    print_fn(
        SummaryTable(
            model_profile.generate_summary(include_macs), title=f"{model.name} summary"
        ).table
    )



import numpy as np
import pytest
import tensorflow as tf

from larq import metrics


def test_config():
    mcv = metrics.FlipRatio(values_dtype="int16", name="mcv", dtype=tf.float16)
    assert mcv.name == "mcv"
    assert mcv.stateful
    assert mcv.dtype == tf.float16
    assert mcv.values_dtype == tf.int16

    mcv2 = metrics.FlipRatio.from_config(mcv.get_config())
    assert mcv2.name == "mcv"
    assert mcv2.stateful
    assert mcv2.dtype == tf.float16
    assert mcv2.values_dtype == tf.int16


@pytest.mark.usefixtures("eager_mode")
def test_metric():
    mcv = metrics.FlipRatio()
    mcv.build((2,))
    assert 0 == mcv.result().numpy()
    assert 0 == mcv.total.numpy()
    assert 0 == mcv.count.numpy()

    mcv.update_state(np.array([1, 1]))
    assert all([1, 1] == mcv._previous_values.numpy())
    assert 0 == mcv.total.numpy()
    assert 1 == mcv.count.numpy()
    assert 0 == mcv.result().numpy()

    mcv.update_state(np.array([2, 2]))
    assert all([2, 2] == mcv._previous_values.numpy())
    assert 1 == mcv.total.numpy()
    assert 2 == mcv.count.numpy()
    assert 1 == mcv.result().numpy()

    mcv.update_state(np.array([1, 2]))
    assert all([1, 2] == mcv._previous_values.numpy())
    assert 1.5 == mcv.total.numpy()
    assert 3 == mcv.count.numpy()
    assert 1.5 / 2 == mcv.result().numpy()


@pytest.mark.usefixtures("eager_mode")
def test_metric_implicit_build():
    mcv = metrics.FlipRatio()

    mcv.update_state(np.array([1, 1]))
    assert all([1, 1] == mcv._previous_values.numpy())
    assert 0 == mcv.total.numpy()
    assert 1 == mcv.count.numpy()
    assert 0 == mcv.result().numpy()

    mcv.update_state(np.array([2, 2]))
    assert all([2, 2] == mcv._previous_values.numpy())
    assert 1 == mcv.total.numpy()
    assert 2 == mcv.count.numpy()
    assert 1 == mcv.result().numpy()

    mcv.update_state(np.array([1, 2]))
    assert all([1, 2] == mcv._previous_values.numpy())
    assert 1.5 == mcv.total.numpy()
    assert 3 == mcv.count.numpy()
    assert 1.5 / 2 == mcv.result().numpy()


@pytest.mark.usefixtures("eager_mode")
def test_metric_wrong_shape():
    mcv = metrics.FlipRatio()
    mcv.build((3,))
    with pytest.raises((ValueError, tf.errors.InvalidArgumentError)):
        mcv.update_state(np.array([1, 1]))


@pytest.mark.usefixtures("graph_mode")
def test_metric_in_graph_mode():
    mcv = metrics.FlipRatio()
    mcv.build((2,))

    new_state = tf.compat.v1.placeholder(dtype=tf.float32, shape=[2])
    update_state_op = mcv.update_state(new_state)
    metric_value = mcv.result()

    with tf.compat.v1.Session() as sess:
        sess.run(tf.compat.v1.variables_initializer(mcv.variables))

        sess.run(update_state_op, feed_dict={new_state: [1, 1]})
        sess.run(update_state_op, feed_dict={new_state: [2, 2]})
        sess.run(update_state_op, feed_dict={new_state: [1, 2]})

        previous, total, count, result = sess.run(
            [mcv._previous_values, mcv.total, mcv.count, metric_value]
        )

    assert all([1, 2] == previous)
    assert 1.5 == total
    assert 3 == count
    assert 1.5 / 2 == result



import logging
from typing import Optional

import tensorflow as tf

from larq import context, quantizers, utils
from larq.quantized_variable import QuantizedVariable
from larq.quantizers import NoOpQuantizer, Quantizer

log = logging.getLogger(__name__)


def _is_binary(quantizer):
    return getattr(quantizer, "precision", None) == 1 and not isinstance(
        quantizer, NoOpQuantizer
    )


def _compute_padded_size(stride, dilation_rate, input_size, filter_size):
    if input_size is None:
        return None
    effective_filter_size = (filter_size - 1) * dilation_rate + 1
    output_size = (input_size + stride - 1) // stride
    padded_size = (output_size - 1) * stride + effective_filter_size
    if tf.is_tensor(input_size):
        return tf.math.maximum(padded_size, input_size)
    return max(padded_size, input_size)


def _compute_padding(stride, dilation_rate, input_size, filter_size):
    padded_size = _compute_padded_size(stride, dilation_rate, input_size, filter_size)
    total_padding = padded_size - input_size
    padding = total_padding // 2
    return padding, padding + (total_padding % 2)


class BaseLayer(tf.keras.layers.Layer):
    """Base class for defining quantized layers.

    `input_quantizer` is the element-wise quantization functions to use.
    If `input_quantizer=None` this layer is equivalent to `tf.keras.layers.Layer`.
    """

    def __init__(self, *args, input_quantizer=None, **kwargs):
        self.input_quantizer = quantizers.get(input_quantizer)
        super().__init__(*args, **kwargs)

    def call(self, inputs):
        if self.input_quantizer:
            inputs = self.input_quantizer(inputs)
        with context.quantized_scope(True):
            return super().call(inputs)

    def get_config(self):
        return {
            **super().get_config(),
            "input_quantizer": quantizers.serialize(self.input_quantizer),
        }

    def _get_quantizer(self, name) -> Optional[Quantizer]:
        """Get quantizer for given kernel name"""
        return None

    def _add_variable_with_custom_getter(self, name: str, **kwargs):
        quantizer = self._get_quantizer(name)
        if quantizer is None:
            return super()._add_variable_with_custom_getter(name, **kwargs)

        old_getter = kwargs.pop("getter")

        # Wrap `getter` with a version that returns a `QuantizedVariable`.
        def getter(*args, **kwargs):
            variable = old_getter(*args, **kwargs)
            return QuantizedVariable.from_variable(variable, quantizer)

        return super()._add_variable_with_custom_getter(name, getter=getter, **kwargs)


class QuantizerBase(BaseLayer):
    """Base class for defining quantized layers with a single kernel.

    `kernel_quantizer` is the element-wise quantization functions to use.
    If `kernel_quantizer=None` this layer is equivalent to `BaseLayer`.
    """

    def __init__(self, *args, kernel_quantizer=None, **kwargs):
        self.kernel_quantizer = quantizers.get_kernel_quantizer(kernel_quantizer)

        super().__init__(*args, **kwargs)
        if _is_binary(self.kernel_quantizer) and not self.kernel_constraint:
            log.warning(
                "Using a binary weight quantizer without setting `kernel_constraint` "
                "may result in starved weights (where the gradient is always zero)."
            )

    def _get_quantizer(self, name: str) -> Optional[Quantizer]:
        return self.kernel_quantizer if name == "kernel" else None

    def get_config(self):
        return {
            **super().get_config(),
            "kernel_quantizer": quantizers.serialize(self.kernel_quantizer),
        }


class QuantizerBaseConv(tf.keras.layers.Layer):
    """Base class for defining quantized conv layers"""

    def __init__(self, *args, pad_values=0.0, **kwargs):
        self.pad_values = pad_values
        super().__init__(*args, **kwargs)
        is_zero_padding = not tf.is_tensor(self.pad_values) and self.pad_values == 0.0
        self._is_native_padding = self.padding != "same" or is_zero_padding
        if self.padding == "causal" and not is_zero_padding:
            raise ValueError("Causal padding with `pad_values != 0` is not supported.")

    def _get_spatial_padding_same(self, shape):
        return [
            _compute_padding(stride, dilation_rate, shape[i], filter_size)
            for i, (stride, dilation_rate, filter_size) in enumerate(
                zip(self.strides, self.dilation_rate, self.kernel_size)
            )
        ]

    def _get_spatial_shape(self, input_shape):
        return (
            input_shape[1:-1]
            if self.data_format == "channels_last"
            else input_shape[2:]
        )

    def _get_padding_same(self, inputs):
        input_shape = inputs.shape
        if not input_shape[1:].is_fully_defined():
            input_shape = tf.shape(inputs)
        padding = self._get_spatial_padding_same(self._get_spatial_shape(input_shape))
        return (
            [[0, 0], *padding, [0, 0]]
            if self.data_format == "channels_last"
            else [[0, 0], [0, 0], *padding]
        )

    def _get_padding_same_shape(self, input_shape):
        spatial_input_shape = self._get_spatial_shape(input_shape)
        spatial_shape = [
            _compute_padded_size(stride, dilation, size, filter_size)
            for size, stride, dilation, filter_size in zip(
                spatial_input_shape,
                self.strides,
                self.dilation_rate,
                self.kernel_size,
            )
        ]
        if self.data_format == "channels_last":
            return tf.TensorShape([input_shape[0], *spatial_shape, input_shape[-1]])
        return tf.TensorShape([*input_shape[:2], *spatial_shape])

    def build(self, input_shape):
        if self._is_native_padding:
            super().build(input_shape)
        else:
            with utils.patch_object(self, "padding", "valid"):
                super().build(self._get_padding_same_shape(input_shape))

    def call(self, inputs):
        if self._is_native_padding:
            return super().call(inputs)

        inputs = tf.pad(
            inputs, self._get_padding_same(inputs), constant_values=self.pad_values
        )
        with utils.patch_object(self, "padding", "valid"):
            return super().call(inputs)

    def get_config(self):
        return {
            **super().get_config(),
            "pad_values": tf.keras.backend.get_value(self.pad_values),
        }


class QuantizerDepthwiseBase(BaseLayer):
    """Base class for defining depthwise quantized layers

    `depthwise_quantizer` is the element-wise quantization functions to use.
    If `depthwise_quantizer=None` this layer is equivalent to `BaseLayer`.
    """

    def __init__(
        self,
        *args,
        depthwise_quantizer: Optional[Quantizer] = None,
        **kwargs,
    ):
        self.depthwise_quantizer = quantizers.get_kernel_quantizer(depthwise_quantizer)

        super().__init__(*args, **kwargs)
        if _is_binary(self.depthwise_quantizer) and not self.depthwise_constraint:
            log.warning(
                "Using a binary weight quantizer without setting `depthwise_constraint` "
                "may result in starved weights (where the gradient is always zero)."
            )

    def _get_quantizer(self, name: str) -> Optional[Quantizer]:
        return self.depthwise_quantizer if name == "depthwise_kernel" else None

    def get_config(self):
        return {
            **super().get_config(),
            "depthwise_quantizer": quantizers.serialize(self.depthwise_quantizer),
        }


class QuantizerSeparableBase(BaseLayer):
    """Base class for defining separable quantized layers.

    `depthwise_quantizer` and `pointwise_quantizer` are the element-wise quantization
    functions to use. If all quantization functions are `None` this layer is equivalent
    to `BaseLayer`.
    """

    def __init__(
        self,
        *args,
        depthwise_quantizer: Optional[Quantizer] = None,
        pointwise_quantizer: Optional[Quantizer] = None,
        **kwargs,
    ):
        self.depthwise_quantizer = quantizers.get_kernel_quantizer(depthwise_quantizer)
        self.pointwise_quantizer = quantizers.get_kernel_quantizer(pointwise_quantizer)

        super().__init__(*args, **kwargs)
        if _is_binary(self.depthwise_quantizer) and not self.depthwise_constraint:
            log.warning(
                "Using a binary `depthwise_quantizer` without setting `depthwise_constraint` "
                "may result in starved weights (where the gradient is always zero)."
            )
        if _is_binary(self.pointwise_quantizer) and not self.pointwise_constraint:
            log.warning(
                "Using a binary `pointwise_quantizer` without setting `pointwise_constraint` "
                "may result in starved weights (where the gradient is always zero)."
            )

    def _get_quantizer(self, name: str) -> Optional[Quantizer]:
        if name == "depthwise_kernel":
            return self.depthwise_quantizer
        if name == "pointwise_kernel":
            return self.pointwise_quantizer
        return None

    def get_config(self):
        return {
            **super().get_config(),
            "depthwise_quantizer": quantizers.serialize(self.depthwise_quantizer),
            "pointwise_quantizer": quantizers.serialize(self.pointwise_quantizer),
        }



import numpy as np
import pytest
import tensorflow as tf

import larq as lq
from larq.testing_utils import generate_real_values_with_zeros


@pytest.mark.parametrize("name", ["weight_clip"])
def test_serialization(name):
    fn = tf.keras.constraints.get(name)
    ref_fn = getattr(lq.constraints, name)()
    assert fn.__class__ == ref_fn.__class__
    config = tf.keras.constraints.serialize(fn)
    fn = tf.keras.constraints.deserialize(config)
    assert fn.__class__ == ref_fn.__class__


def test_clip():
    real_values = generate_real_values_with_zeros()
    clip_instance = lq.constraints.weight_clip(clip_value=0.75)
    result = clip_instance(tf.keras.backend.variable(real_values))
    result = tf.keras.backend.eval(result)
    np.testing.assert_allclose(result, np.clip(real_values, -0.75, 0.75))



import numpy as np
import tensorflow as tf

import larq as lq


def _eval_tensor(tensor):
    if tensor is None:
        return None
    elif callable(tensor):
        return _eval_helper(tensor())
    else:
        return tensor.numpy()


def _eval_helper(tensors):
    if tensors is None:
        return None
    return tf.nest.map_structure(_eval_tensor, tensors)


def evaluate(tensors):
    if tf.executing_eagerly():
        return _eval_helper(tensors)
    else:
        sess = tf.compat.v1.get_default_session()
        return sess.run(tensors)


def generate_real_values_with_zeros(low=-2, high=2, shape=(4, 10)):
    real_values = np.random.uniform(low, high, shape)
    real_values = np.insert(real_values, 1, 0, axis=1)
    return real_values


def get_small_bnn_model(input_dim, num_hidden, output_dim, trainable_bn=True):
    model = tf.keras.models.Sequential()
    model.add(
        lq.layers.QuantDense(
            units=num_hidden,
            kernel_quantizer="ste_sign",
            kernel_constraint="weight_clip",
            activation="relu",
            input_shape=(input_dim,),
            use_bias=False,
        )
    )
    model.add(tf.keras.layers.BatchNormalization(trainable=trainable_bn))
    model.add(
        lq.layers.QuantDense(
            units=output_dim,
            kernel_quantizer="ste_sign",
            kernel_constraint="weight_clip",
            input_quantizer="ste_sign",
            activation="softmax",
            use_bias=False,
        )
    )
    return model


def random_input(shape):
    for i, dim in enumerate(shape):
        if dim is None:
            shape[i] = np.random.randint(1, 4)
    data = 10 * np.random.random(shape) - 0.5
    return data.astype("float32")


# This is a fork of https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/keras/testing_utils.py#L72
# as recommended in https://github.com/tensorflow/tensorflow/issues/28601#issuecomment-492810252
def layer_test(
    layer_cls,
    kwargs=None,
    input_shape=None,
    input_dtype=None,
    input_data=None,
    expected_output=None,
    expected_output_dtype=None,
    should_run_eagerly=False,
):
    """Test routine for a layer with a single input and single output.
    Arguments:
      layer_cls: Layer class object.
      kwargs: Optional dictionary of keyword arguments for instantiating the
        layer.
      input_shape: Input shape tuple.
      input_dtype: Data type of the input data.
      input_data: Numpy array of input data.
      expected_output: Shape tuple for the expected shape of the output.
      expected_output_dtype: Data type expected for the output.
    Returns:
      The output data (Numpy array) returned by the layer, for additional
      checks to be done by the calling code.
    Raises:
      ValueError: if `input_shape is None`.
    """
    if input_data is None:
        if input_shape is None:
            raise ValueError("input_shape is None")
        if not input_dtype:
            input_dtype = "float32"
        input_data_shape = list(input_shape)
        for i, e in enumerate(input_data_shape):
            if e is None:
                input_data_shape[i] = np.random.randint(1, 4)
        input_data = 10 * np.random.random(input_data_shape)
        if input_dtype[:5] == "float":
            input_data -= 0.5
        input_data = input_data.astype(input_dtype)
    elif input_shape is None:
        input_shape = input_data.shape
    if input_dtype is None:
        input_dtype = input_data.dtype
    if expected_output_dtype is None:
        expected_output_dtype = input_dtype

    # instantiation
    kwargs = kwargs or {}
    layer = layer_cls(**kwargs)

    # test get_weights , set_weights at layer level
    weights = layer.get_weights()
    layer.set_weights(weights)

    # test in functional API
    x = tf.keras.layers.Input(shape=input_shape[1:], dtype=input_dtype)
    y = layer(x)
    if tf.keras.backend.dtype(y) != expected_output_dtype:
        raise AssertionError(
            "When testing layer %s, for input %s, found output "
            "dtype=%s but expected to find %s.\nFull kwargs: %s"
            % (
                layer_cls.__name__,
                x,
                tf.keras.backend.dtype(y),
                expected_output_dtype,
                kwargs,
            )
        )
    # check shape inference
    model = tf.keras.models.Model(x, y)
    expected_output_shape = tuple(
        layer.compute_output_shape(tf.TensorShape(input_shape)).as_list()
    )
    actual_output = model.predict(input_data)
    actual_output_shape = actual_output.shape
    for expected_dim, actual_dim in zip(expected_output_shape, actual_output_shape):
        if expected_dim is not None:
            if expected_dim != actual_dim:
                raise AssertionError(
                    "When testing layer %s, for input %s, found output_shape="
                    "%s but expected to find %s.\nFull kwargs: %s"
                    % (
                        layer_cls.__name__,
                        x,
                        actual_output_shape,
                        expected_output_shape,
                        kwargs,
                    )
                )
    if expected_output is not None:
        np.testing.assert_allclose(actual_output, expected_output, rtol=1e-3)

    # test serialization, weight setting at model level
    model_config = model.get_config()
    recovered_model = tf.keras.models.Model.from_config(model_config)
    if model.weights:
        weights = model.get_weights()
        recovered_model.set_weights(weights)
        output = recovered_model.predict(input_data)
        np.testing.assert_allclose(output, actual_output, rtol=2e-3)

    # Recreate layer to prevent layer metrics from being configured multiple times.
    layer = layer_cls(**kwargs)
    # test training mode (e.g. useful for dropout tests)
    # Rebuild the model to avoid the graph being reused between predict() and
    # train(). This was causing some error for layer with Defun as it body.
    # See b/120160788 for more details. This should be mitigated after 2.0.
    model = tf.keras.models.Model(x, layer(x))
    model.compile(
        "rmsprop",
        "mse",
        weighted_metrics=["acc"],
        run_eagerly=should_run_eagerly,
    )
    model.train_on_batch(input_data, actual_output)

    # test as first layer in Sequential API
    layer_config = layer.get_config()
    layer_config["batch_input_shape"] = input_shape
    layer = layer.__class__.from_config(layer_config)

    model = tf.keras.models.Sequential()
    model.add(layer)
    actual_output = model.predict(input_data)
    actual_output_shape = actual_output.shape
    for expected_dim, actual_dim in zip(expected_output_shape, actual_output_shape):
        if expected_dim is not None:
            if expected_dim != actual_dim:
                raise AssertionError(
                    "When testing layer %s **after deserialization**, "
                    "for input %s, found output_shape="
                    "%s but expected to find inferred shape %s.\nFull kwargs: %s"
                    % (
                        layer_cls.__name__,
                        x,
                        actual_output_shape,
                        expected_output_shape,
                        kwargs,
                    )
                )
    if expected_output is not None:
        np.testing.assert_allclose(actual_output, expected_output, rtol=1e-3)

    # test serialization, weight setting at model level
    model_config = model.get_config()
    recovered_model = tf.keras.models.Sequential.from_config(model_config)
    if model.weights:
        weights = model.get_weights()
        recovered_model.set_weights(weights)
        output = recovered_model.predict(input_data)
        np.testing.assert_allclose(output, actual_output, rtol=2e-3)

    # for further checks in the caller function
    return actual_output



import numpy as np
import pytest
import tensorflow as tf

import larq as lq
from larq.testing_utils import generate_real_values_with_zeros


@pytest.mark.parametrize("fn", [lq.math.sign])
def test_sign(fn):
    x = tf.keras.backend.placeholder(ndim=2)
    f = tf.keras.backend.function([x], [fn(x)])
    binarized_values = np.random.choice([-1, 1], size=(2, 5)).astype(np.float32)
    result = f(binarized_values)[0]
    np.testing.assert_allclose(result, binarized_values)

    real_values = generate_real_values_with_zeros()
    result = f(real_values)[0]
    assert not np.any(result == 0)
    assert np.all(result[real_values < 0] == -1)
    assert np.all(result[real_values >= 0] == 1)

    zero_values = np.zeros((2, 5))
    result = f(zero_values)[0]
    assert np.all(result == 1)


@pytest.mark.parametrize("fn", [lq.math.heaviside])
def test_heaviside(fn):
    x = tf.keras.backend.placeholder(ndim=2)
    f = tf.keras.backend.function([x], [fn(x)])
    binarized_values = np.random.choice([0, 1], size=(2, 5))
    result = f([binarized_values])[0]
    np.testing.assert_allclose(result, binarized_values)

    real_values = generate_real_values_with_zeros()
    result = f([real_values])[0]
    assert np.all(result[real_values <= 0] == 0)
    assert np.all(result[real_values > 0] == 1)



"""Functions from the `constraints` module allow setting constraints
(eg. weight clipping) on network parameters during optimization.

The penalties are applied on a per-layer basis. The exact API will depend on the layer,
but the layers `QuantDense`, `QuantConv1D`, `QuantConv2D` and `QuantConv3D` have a
unified API.

These layers expose 2 keyword arguments:

- `kernel_constraint` for the main weights matrix
- `bias_constraint` for the bias.

```python
import larq as lq

lq.layers.QuantDense(64, kernel_constraint="weight_clip")
lq.layers.QuantDense(64, kernel_constraint=lq.constraints.WeightClip(2.))
```
"""

from typing import Any, Mapping

import tensorflow as tf

from larq import utils


@utils.register_keras_custom_object
class WeightClip(tf.keras.constraints.Constraint):
    """Weight Clip constraint

    Constrains the weights incident to each hidden unit
    to be between `[-clip_value, clip_value]`.

    # Arguments
        clip_value: The value to clip incoming weights.
    """

    def __init__(self, clip_value: float = 1):
        self.clip_value = clip_value

    def __call__(self, x: tf.Tensor) -> tf.Tensor:
        return tf.clip_by_value(x, -self.clip_value, self.clip_value)

    def get_config(self) -> Mapping[str, Any]:
        return {"clip_value": self.clip_value}


# Aliases
@utils.register_keras_custom_object
class weight_clip(WeightClip):
    pass



from larq import (
    activations,
    callbacks,
    constraints,
    context,
    layers,
    math,
    metrics,
    models,
    optimizers,
    quantizers,
    utils,
)

try:
    from importlib import metadata  # type: ignore
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata  # type: ignore

__version__ = metadata.version("larq")

__all__ = [
    "layers",
    "activations",
    "callbacks",
    "constraints",
    "context",
    "math",
    "metrics",
    "models",
    "quantizers",
    "optimizers",
    "utils",
]



from contextlib import contextmanager

from tensorflow.keras.utils import get_custom_objects


def memory_as_readable_str(num_bits: int) -> str:
    """Generate a human-readable string for the memory size.

    1 KiB = 1024 B; we use the binary prefix (KiB) [1,2] instead of the decimal prefix
    (KB) to avoid any confusion with multiplying by 1000 instead of 1024.

    [1] https://en.wikipedia.org/wiki/Binary_prefix
    [2] https://physics.nist.gov/cuu/Units/binary.html
    """

    suffixes = ["B", "KiB", "MiB", "GiB"]
    num_bytes = num_bits / 8

    for i, suffix in enumerate(suffixes):
        rounded = num_bytes / (1024 ** i)
        if rounded < 1024:
            break

    return f"{rounded:,.2f} {suffix}"


def register_keras_custom_object(cls):
    """See https://github.com/tensorflow/addons/blob/master/tensorflow_addons/utils/keras_utils.py#L25"""
    get_custom_objects()[cls.__name__] = cls
    return cls


def register_alias(name: str):
    """A decorator to register a custom keras object under a given alias.
    !!! example
        ```python
        @utils.register_alias("degeneration")
        class Degeneration(tf.keras.metrics.Metric):
            pass
        ```
    """

    def register_func(cls):
        get_custom_objects()[name] = cls
        return cls

    return register_func


def set_precision(precision: int = 32):
    """A decorator to set the precision of a quantizer function

    # Arguments
        precision: An integer defining the precision of the output.
    """

    def decorator(function):
        setattr(function, "precision", precision)
        return function

    return decorator


@contextmanager
def patch_object(object, name, value):
    """Temporarily overwrite attribute on object"""
    old_value = getattr(object, name)
    setattr(object, name, value)
    yield
    setattr(object, name, old_value)



"""A Quantizer defines the way of transforming a full precision input to a
quantized output and the pseudo-gradient method used for the backwards pass.

Quantizers can either be used through quantizer arguments that are supported
for Larq layers, such as `input_quantizer` and `kernel_quantizer`; or they
can be used similar to activations, i.e. either through an `Activation` layer,
or through the `activation` argument supported by all forward layers:

```python
import tensorflow as tf
import larq as lq
...
x = lq.layers.QuantDense(64, activation=None)(x)
x = lq.layers.QuantDense(64, input_quantizer="ste_sign")(x)
```

is equivalent to:

```python
x = lq.layers.QuantDense(64)(x)
x = tf.keras.layers.Activation("ste_sign")(x)
x = lq.layers.QuantDense(64)(x)
```

as well as:

```python
x = lq.layers.QuantDense(64, activation="ste_sign")(x)
x = lq.layers.QuantDense(64)(x)
```

We highly recommend using the first of these formulations: for the
other two formulations, intermediate layers - like batch normalization or
average pooling - and shortcut connections may result in non-binary input
to the convolutions.

Quantizers can either be referenced by string or called directly.
The following usages are equivalent:

```python
lq.layers.QuantDense(64, kernel_quantizer="ste_sign")
```
```python
lq.layers.QuantDense(64, kernel_quantizer=lq.quantizers.SteSign(clip_value=1.0))
```
"""
from typing import Callable, Union

import tensorflow as tf

from larq import context, math
from larq import metrics as lq_metrics
from larq import utils

__all__ = [
    "ApproxSign",
    "DoReFa",
    "DoReFaQuantizer",
    "MagnitudeAwareSign",
    "NoOp",
    "NoOpQuantizer",
    "Quantizer",
    "SteHeaviside",
    "SteSign",
    "SteTern",
    "SwishSign",
]


def _clipped_gradient(x, dy, clip_value):
    """Calculate `clipped_gradent * dy`."""

    if clip_value is None:
        return dy

    zeros = tf.zeros_like(dy)
    mask = tf.math.less_equal(tf.math.abs(x), clip_value)
    return tf.where(mask, dy, zeros)


def ste_sign(x: tf.Tensor, clip_value: float = 1.0) -> tf.Tensor:
    @tf.custom_gradient
    def _call(x):
        def grad(dy):
            return _clipped_gradient(x, dy, clip_value)

        return math.sign(x), grad

    return _call(x)


def _scaled_sign(x):  # pragma: no cover
    return 1.3 * ste_sign(x)


@tf.custom_gradient
def approx_sign(x: tf.Tensor) -> tf.Tensor:
    def grad(dy):
        abs_x = tf.math.abs(x)
        zeros = tf.zeros_like(dy)
        mask = tf.math.less_equal(abs_x, 1.0)
        return tf.where(mask, (1 - abs_x) * 2 * dy, zeros)

    return math.sign(x), grad


def swish_sign(x: tf.Tensor, beta: float = 5.0) -> tf.Tensor:
    @tf.custom_gradient
    def _call(x):
        def grad(dy):
            b_x = beta * x
            return dy * beta * (2 - b_x * tf.tanh(b_x * 0.5)) / (1 + tf.cosh(b_x))

        return math.sign(x), grad

    return _call(x)


def ste_tern(
    x: tf.Tensor,
    threshold_value: float = 0.05,
    ternary_weight_networks: bool = False,
    clip_value: float = 1.0,
) -> tf.Tensor:
    @tf.custom_gradient
    def _call(x):
        if ternary_weight_networks:
            threshold = 0.7 * tf.reduce_sum(tf.abs(x)) / tf.cast(tf.size(x), x.dtype)
        else:
            threshold = threshold_value

        def grad(dy):
            return _clipped_gradient(x, dy, clip_value)

        return tf.sign(tf.sign(x + threshold) + tf.sign(x - threshold)), grad

    return _call(x)


def ste_heaviside(x: tf.Tensor, clip_value: float = 1.0) -> tf.Tensor:
    @tf.custom_gradient
    def _call(x):
        def grad(dy):
            return _clipped_gradient(x, dy, clip_value)

        return math.heaviside(x), grad

    return _call(x)


class Quantizer(tf.keras.layers.Layer):
    """Common base class for defining quantizers.

    # Attributes
        precision: An integer defining the precision of the output. This value will be
            used by `lq.models.summary()` for improved logging.
    """

    precision = None

    def compute_output_shape(self, input_shape):
        return input_shape


class _BaseQuantizer(Quantizer):
    """Private base class for defining quantizers with Larq metrics."""

    def __init__(self, *args, metrics=None, **kwargs):
        self._custom_metrics = metrics
        super().__init__(*args, **kwargs)

    def build(self, input_shape):
        if self._custom_metrics and "flip_ratio" in self._custom_metrics:
            self.flip_ratio = lq_metrics.FlipRatio(name=f"flip_ratio/{self.name}")
            self.flip_ratio.build(input_shape)
        super().build(input_shape)

    def call(self, inputs):
        if hasattr(self, "flip_ratio"):
            self.add_metric(self.flip_ratio(inputs))
        return inputs

    @property
    def non_trainable_weights(self):
        return []


@utils.register_keras_custom_object
class NoOp(_BaseQuantizer):
    r"""Instantiates a serializable no-op quantizer.

    \\[
    q(x) = x
    \\]

    !!! warning
        This quantizer will not change the input variable. It is only intended to mark
        variables with a desired precision that will be recognized by optimizers like
        `Bop` and add training metrics to track variable changes.

    !!! example
        ```python
        layer = lq.layers.QuantDense(
            16, kernel_quantizer=lq.quantizers.NoOpQuantizer(precision=1),
        )
        layer.build((32,))
        assert layer.kernel.precision == 1
        ```

    # Arguments
        precision: Set the desired precision of the variable. This can be used to tag
        metrics: An array of metrics to add to the layer. If `None` the metrics set in
            `larq.context.metrics_scope` are used. Currently only the `flip_ratio`
            metric is available.
    """
    precision = None

    def __init__(self, precision: int, **kwargs):
        self.precision = precision
        super().__init__(**kwargs)

    def get_config(self):
        return {**super().get_config(), "precision": self.precision}


# `NoOp` used to be called `NoOpQuantizer`; this alias is for
# backwards-compatibility.
NoOpQuantizer = NoOp


@utils.register_alias("ste_sign")
@utils.register_keras_custom_object
class SteSign(_BaseQuantizer):
    r"""Instantiates a serializable binary quantizer.

    \\[
    q(x) = \begin{cases}
      -1 & x < 0 \\\
      1 & x \geq 0
    \end{cases}
    \\]

    The gradient is estimated using the Straight-Through Estimator
    (essentially the binarization is replaced by a clipped identity on the
    backward pass).
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
      1 & \left|x\right| \leq \texttt{clip_value} \\\
      0 & \left|x\right| > \texttt{clip_value}
    \end{cases}\\]

    ```plot-activation
    quantizers.SteSign
    ```

    # Arguments
        clip_value: Threshold for clipping gradients. If `None` gradients are not
            clipped.
        metrics: An array of metrics to add to the layer. If `None` the metrics set in
            `larq.context.metrics_scope` are used. Currently only the `flip_ratio`
            metric is available.

    # References
        - [Binarized Neural Networks: Training Deep Neural Networks with Weights and
            Activations Constrained to +1 or -1](https://arxiv.org/abs/1602.02830)
    """
    precision = 1

    def __init__(self, clip_value: float = 1.0, **kwargs):
        self.clip_value = clip_value
        super().__init__(**kwargs)

    def call(self, inputs):
        outputs = ste_sign(inputs, clip_value=self.clip_value)
        return super().call(outputs)

    def get_config(self):
        return {**super().get_config(), "clip_value": self.clip_value}


@utils.register_alias("approx_sign")
@utils.register_keras_custom_object
class ApproxSign(_BaseQuantizer):
    r"""Instantiates a serializable binary quantizer.
    \\[
    q(x) = \begin{cases}
      -1 & x < 0 \\\
      1 & x \geq 0
    \end{cases}
    \\]

    The gradient is estimated using the ApproxSign method.
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
      (2 - 2 \left|x\right|) & \left|x\right| \leq 1 \\\
      0 & \left|x\right| > 1
    \end{cases}
    \\]

    ```plot-activation
    quantizers.ApproxSign
    ```

    # Arguments
        metrics: An array of metrics to add to the layer. If `None` the metrics set in
            `larq.context.metrics_scope` are used. Currently only the `flip_ratio`
            metric is available.

    # References
        - [Bi-Real Net: Enhancing the Performance of 1-bit CNNs With Improved
            Representational Capability and Advanced Training
            Algorithm](https://arxiv.org/abs/1808.00278)
    """
    precision = 1

    def call(self, inputs):
        outputs = approx_sign(inputs)
        return super().call(outputs)


@utils.register_alias("ste_heaviside")
@utils.register_keras_custom_object
class SteHeaviside(_BaseQuantizer):
    r"""
    Instantiates a binarization quantizer with output values 0 and 1.
    \\[
    q(x) = \begin{cases}
    +1 & x > 0 \\\
    0 & x \leq 0
    \end{cases}
    \\]

    The gradient is estimated using the Straight-Through Estimator
    (essentially the binarization is replaced by a clipped identity on the
    backward pass).

    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
    1 & \left|x\right| \leq 1 \\\
    0 & \left|x\right| > 1
    \end{cases}\\]

    ```plot-activation
    quantizers.SteHeaviside
    ```

    # Arguments
        clip_value: Threshold for clipping gradients. If `None` gradients are not
            clipped.
        metrics: An array of metrics to add to the layer. If `None` the metrics set in
            `larq.context.metrics_scope` are used. Currently only the `flip_ratio`
            metric is available.

    # Returns
        AND Binarization function
    """
    precision = 1

    def __init__(self, clip_value: float = 1.0, **kwargs):
        self.clip_value = clip_value
        super().__init__(**kwargs)

    def call(self, inputs):
        outputs = ste_heaviside(inputs, clip_value=self.clip_value)
        return super().call(outputs)

    def get_config(self):
        return {**super().get_config(), "clip_value": self.clip_value}


@utils.register_alias("swish_sign")
@utils.register_keras_custom_object
class SwishSign(_BaseQuantizer):
    r"""Sign binarization function.

    \\[
    q(x) = \begin{cases}
      -1 & x < 0 \\\
      1 & x \geq 0
    \end{cases}
    \\]

    The gradient is estimated using the SignSwish method.

    \\[
    \frac{\partial q_{\beta}(x)}{\partial x} = \frac{\beta\left\\{2-\beta x \tanh \left(\frac{\beta x}{2}\right)\right\\}}{1+\cosh (\beta x)}
    \\]

    ```plot-activation
    quantizers.SwishSign
    ```
    # Arguments
        beta: Larger values result in a closer approximation to the derivative of the
            sign.
        metrics: An array of metrics to add to the layer. If `None` the metrics set in
            `larq.context.metrics_scope` are used. Currently only the `flip_ratio`
            metric is available.

    # Returns
        SwishSign quantization function

    # References
        - [BNN+: Improved Binary Network Training](https://arxiv.org/abs/1812.11800)
    """
    precision = 1

    def __init__(self, beta: float = 5.0, **kwargs):
        self.beta = beta
        super().__init__(**kwargs)

    def call(self, inputs):
        outputs = swish_sign(inputs, beta=self.beta)
        return super().call(outputs)

    def get_config(self):
        return {**super().get_config(), "beta": self.beta}


@utils.register_alias("magnitude_aware_sign")
@utils.register_keras_custom_object
class MagnitudeAwareSign(_BaseQuantizer):
    r"""Instantiates a serializable magnitude-aware sign quantizer for Bi-Real Net.

    A scaled sign function computed according to Section 3.3 in
    [Zechun Liu et al](https://arxiv.org/abs/1808.00278).

    ```plot-activation
    quantizers._scaled_sign
    ```

    # Arguments
        clip_value: Threshold for clipping gradients. If `None` gradients are not
            clipped.
        metrics: An array of metrics to add to the layer. If `None` the metrics set in
            `larq.context.metrics_scope` are used. Currently only the `flip_ratio`
            metric is available.

    # References
        - [Bi-Real Net: Enhancing the Performance of 1-bit CNNs With Improved
        Representational Capability and Advanced Training
        Algorithm](https://arxiv.org/abs/1808.00278)

    """
    precision = 1

    def __init__(self, clip_value: float = 1.0, **kwargs):
        self.clip_value = clip_value
        super().__init__(**kwargs)

    def call(self, inputs):
        scale_factor = tf.stop_gradient(
            tf.reduce_mean(tf.abs(inputs), axis=list(range(len(inputs.shape) - 1)))
        )

        outputs = scale_factor * ste_sign(inputs, clip_value=self.clip_value)
        return super().call(outputs)

    def get_config(self):
        return {**super().get_config(), "clip_value": self.clip_value}


@utils.register_alias("ste_tern")
@utils.register_keras_custom_object
class SteTern(_BaseQuantizer):
    r"""Instantiates a serializable ternarization quantizer.

    \\[
    q(x) = \begin{cases}
    +1 & x > \Delta \\\
    0 & |x| < \Delta \\\
     -1 & x < - \Delta
    \end{cases}
    \\]

    where \\(\Delta\\) is defined as the threshold and can be passed as an argument,
    or can be calculated as per the Ternary Weight Networks original paper, such that

    \\[
    \Delta = \frac{0.7}{n} \sum_{i=1}^{n} |W_i|
    \\]
    where we assume that \\(W_i\\) is generated from a normal distribution.

    The gradient is estimated using the Straight-Through Estimator
    (essentially the Ternarization is replaced by a clipped identity on the
    backward pass).
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
    1 & \left|x\right| \leq \texttt{clip_value} \\\
    0 & \left|x\right| > \texttt{clip_value}
    \end{cases}\\]

    ```plot-activation
    quantizers.SteTern
    ```

    # Arguments
        threshold_value: The value for the threshold, \\(\Delta\\).
        ternary_weight_networks: Boolean of whether to use the
            Ternary Weight Networks threshold calculation.
        clip_value: Threshold for clipping gradients. If `None` gradients are not
            clipped.
        metrics: An array of metrics to add to the layer. If `None` the metrics set in
            `larq.context.metrics_scope` are used. Currently only the `flip_ratio`
            metric is available.

    # References
        - [Ternary Weight Networks](https://arxiv.org/abs/1605.04711)
    """

    precision = 2

    def __init__(
        self,
        threshold_value: float = 0.05,
        ternary_weight_networks: bool = False,
        clip_value: float = 1.0,
        **kwargs,
    ):
        self.threshold_value = threshold_value
        self.ternary_weight_networks = ternary_weight_networks
        self.clip_value = clip_value
        super().__init__(**kwargs)

    def call(self, inputs):
        outputs = ste_tern(
            inputs,
            threshold_value=self.threshold_value,
            ternary_weight_networks=self.ternary_weight_networks,
            clip_value=self.clip_value,
        )
        return super().call(outputs)

    def get_config(self):
        return {
            **super().get_config(),
            "threshold_value": self.threshold_value,
            "ternary_weight_networks": self.ternary_weight_networks,
            "clip_value": self.clip_value,
        }


@utils.register_alias("dorefa_quantizer")
@utils.register_keras_custom_object
class DoReFa(_BaseQuantizer):
    r"""Instantiates a serializable k_bit quantizer as in the DoReFa paper.

    \\[
    q(x) = \begin{cases}
    0 & x < \frac{1}{2n} \\\
    \frac{i}{n} & \frac{2i-1}{2n} < x < \frac{2i+1}{2n} \text{ for } i \in \\{1,n-1\\}\\\
     1 & \frac{2n-1}{2n} < x
    \end{cases}
    \\]

    where \\(n = 2^{\text{k_bit}} - 1\\). The number of bits, k_bit, needs to be passed
    as an argument.
    The gradient is estimated using the Straight-Through Estimator
    (essentially the binarization is replaced by a clipped identity on the
    backward pass).
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
    1 &  0 \leq x \leq 1 \\\
    0 & \text{else}
    \end{cases}\\]

    !!! warning
        While the DoReFa paper describes how to do quantization for both weights and
        activations, this implementation is only valid for activations, and this
        quantizer should therefore not be used as a kernel quantizer.

    ```plot-activation
    quantizers.DoReFa
    ```

    # Arguments
        k_bit: number of bits for the quantization.
        metrics: An array of metrics to add to the layer. If `None` the metrics set in
            `larq.context.metrics_scope` are used. Currently only the `flip_ratio`
            metric is available.

    # Returns
        Quantization function

    # References
        - [DoReFa-Net: Training Low Bitwidth Convolutional Neural Networks with Low
            Bitwidth Gradients](https://arxiv.org/abs/1606.06160)
    """
    precision = None

    def __init__(self, k_bit: int = 2, **kwargs):
        self.precision = k_bit
        super().__init__(**kwargs)

    def call(self, inputs):
        inputs = tf.clip_by_value(inputs, 0.0, 1.0)

        @tf.custom_gradient
        def _k_bit_with_identity_grad(x):
            n = 2 ** self.precision - 1
            return tf.round(x * n) / n, lambda dy: dy

        outputs = _k_bit_with_identity_grad(inputs)
        return super().call(outputs)

    def get_config(self):
        return {**super().get_config(), "k_bit": self.precision}


# `DoReFa` used to be called `DoReFaQuantizer`; this alias is for
# backwards-compatibility.
DoReFaQuantizer = DoReFa


Quantizer = Union[tf.keras.layers.Layer, Callable[[tf.Tensor], tf.Tensor]]


def serialize(quantizer: tf.keras.layers.Layer):
    return tf.keras.utils.serialize_keras_object(quantizer)


def deserialize(name, custom_objects=None):
    return tf.keras.utils.deserialize_keras_object(
        name,
        module_objects=globals(),
        custom_objects=custom_objects,
        printable_module_name="quantization function",
    )


def get(identifier):
    if identifier is None:
        return None
    if isinstance(identifier, dict):
        return deserialize(identifier)
    if isinstance(identifier, str):
        return deserialize(str(identifier))
    if callable(identifier):
        return identifier
    raise ValueError(
        f"Could not interpret quantization function identifier: {identifier}"
    )


def get_kernel_quantizer(identifier):
    """Returns a quantizer from identifier and adds default kernel quantizer metrics.

    # Arguments
        identifier: Function or string

    # Returns
        `Quantizer` or `None`
    """
    quantizer = get(identifier)
    if isinstance(quantizer, _BaseQuantizer) and not quantizer._custom_metrics:
        quantizer._custom_metrics = list(context.get_training_metrics())
    return quantizer



"""Activations can either be used through an `Activation` layer, or through the
`activation` argument supported by all forward layers:

```python
import tensorflow as tf
import larq as lq

model.add(lq.layers.QuantDense(64))
model.add(tf.keras.layers.Activation('hard_tanh'))
```

This is equivalent to:

```python
model.add(lq.layers.QuantDense(64, activation='hard_tanh'))
```

You can also pass an element-wise TensorFlow function as an activation:

```python
model.add(lq.layers.QuantDense(64, activation=lq.activations.hard_tanh))
```
"""

import tensorflow as tf

from larq import utils


@utils.register_keras_custom_object
def hard_tanh(x: tf.Tensor) -> tf.Tensor:
    """Hard tanh activation function.
    ```plot-activation
    activations.hard_tanh
    ```

    # Arguments
        x: Input tensor.

    # Returns
        Hard tanh activation.
    """
    return tf.clip_by_value(x, -1, 1)


@utils.register_keras_custom_object
def leaky_tanh(x: tf.Tensor, alpha: float = 0.2) -> tf.Tensor:
    r"""Leaky tanh activation function.
    Similar to hard tanh, but with non-zero slopes as in leaky ReLU.

    ```plot-activation
    activations.leaky_tanh
    ```

    # Arguments
        x: Input tensor.
        alpha: Slope of the activation function outside of [-1, 1].

    # Returns
        Leaky tanh activation.
    """
    return (
        tf.clip_by_value(x, -1, 1)
        + (tf.math.maximum(x, 1) - 1) * alpha
        + (tf.math.minimum(x, -1) + 1) * alpha
    )



"""Neural networks with extremely low-precision weights and activations, such as
Binarized Neural Networks (BNNs), usually contain a mix of low-precision weights (e.g.
1-bit) and  higher-precision weights (e.g. 8-bit, 16-bit, or 32-bit). Examples of this
include the first and last layers of image classificiation models, which have
higher-precision weights in most BNN architectures from the literature.

Training a BNN, then, consists of optimizing both low-precision and higher-precision
weights. In `larq`, we provide a mechanism to target different bit-precision variables
with different optimizers using the `CaseOptimizer` class. Modeled after the
[`tf.case`](https://www.tensorflow.org/api_docs/python/tf/case) signature,
`CaseOptimizer` accepts pairs of predicates and optimizers. A predicate, given a
variable, decides whether its optimizer should train that variable.

A `CaseOptimizer` behaves much like any other
[Keras optimizer](https://www.tensorflow.org/api_docs/python/tf/keras/optimizers), and
once you instantiate it you can pass it to your `model.compile()` as usual. To
instantiate a `CaseOptimzer`, pass one or a list of `(predicate, optimizer)` tuples,
along with a `default` optimizer which trains any variables not claimed by another
optimizer. A variable may not be claimed by more than one optimizer's predicate.

!!! example
    ```python
    no_op_quantizer = lq.quantizers.NoOp(precision=1)
    layer = lq.layers.QuantDense(16, kernel_quantizer=no_op_quantizer)

    case_optimizer = lq.optimizers.CaseOptimizer(
        (
            lq.optimizers.Bop.is_binary_variable,  # predicate
            lq.optimizers.Bop(threshold=1e-6, gamma=1e-3),  # optimizer
        ),
        default_optimizer=tf.keras.optimizers.Adam(0.01),
    )
    ```
"""


import warnings
from copy import deepcopy
from typing import Callable, Optional, Tuple

import tensorflow as tf

import larq as lq
from larq import utils

__all__ = ["Bop", "CaseOptimizer"]


@utils.register_keras_custom_object
class CaseOptimizer(tf.keras.optimizers.Optimizer):
    """An optmizer wrapper that applies different optimizers to a subset of variables.

    An optimizer is used to train a variable iff its accompanying predicate evaluates to
    `True`.

    For each variable, at most one optimizer's predicate may evaluate to `True`. If no
    optimizer's predicate evaluates to `True` for a variable, it is trained with the
    `default_optimizer`. If a variable is claimed by no optimizers and
    `default_optimizer == None`, the variable is not trained.

    # Arguments
        predicate_optimizer_pairs: One or more `(pred, tf.keras.optimizers.Optimizer)`
            pairs, where `pred`  takes one `tf.Variable` as argument and returns `True`
            if the optimizer should be used for that variable, e.g. `pred(var) == True`.
        default_optimizer: A `tf.keras.optimizers.Optimizer` to be applied to any
            variable not claimed by any other optimizer. (Must be passed as keyword
            argument.)
    """

    _HAS_AGGREGATE_GRAD = True

    def __init__(
        self,
        *predicate_optimizer_pairs: Tuple[
            Callable[[tf.Variable], bool], tf.keras.optimizers.Optimizer
        ],
        default_optimizer: Optional[tf.keras.optimizers.Optimizer] = None,
        name: str = "optimizer_case",
    ):
        super().__init__(name=name)

        # Type checks for (predicate, optimizer) pairs
        for i, (predicate, optimizer) in enumerate(predicate_optimizer_pairs):
            if not callable(predicate):
                raise TypeError(
                    f"Expected callable predicate at `predicate_optimizer_pairs[{i}][0]` but got `{type(predicate)}`."
                )
            if not isinstance(optimizer, tf.keras.optimizers.Optimizer):
                raise TypeError(
                    f"Expected `tf.keras.optimizers.Optimizer` at `predicate_optimizer_pairs[{i}][1]` but got `{type(optimizer)}`."
                )

        # Type check for default optimizers
        if default_optimizer is not None and not isinstance(
            default_optimizer, tf.keras.optimizers.Optimizer
        ):
            raise TypeError(
                f"Expected `tf.keras.optimizers.Optimizer` for `default_optimizer` but got `{type(default_optimizer)}`."
            )

        self.pred_opt_pairs = predicate_optimizer_pairs
        self.default = default_optimizer

        self.var_opt_mapping = None

        # List of optimizers ending in `default_optimizer`, for easier internal access
        self.optimizers = [opt for (_, opt) in self.pred_opt_pairs]

        if self.default:
            self.optimizers.append(self.default)
            self.DEFAULT_OPT_INDEX = len(self.pred_opt_pairs)

        # Track optimizers to support reloading via tf.train.Checkpoint
        for i, optimizer in enumerate(self.optimizers):
            self._track_trackable(optimizer, name=f"optimizer_{i}")

    @property
    def weights(self):
        weights = []
        for optimizer in self.optimizers:
            weights.extend(optimizer.weights)
        return weights

    def apply_gradients(self, grads_and_vars, name: Optional[str] = None, **kwargs):
        """Apply gradients to variables for each optimizer.

        On the first call to `apply_gradients()`, compute the mapping from variables to
        optimizers and cache it in the `self.var_opt_mapping` dict for serialization and
        faster access.
        """

        if self.var_opt_mapping is None:
            # Convert `grads_and_vars` to list so we can iterate multiple times over it
            grads_and_vars = list(grads_and_vars)
            self._compute_var_opt_mapping(grads_and_vars)

        # Split gradients and variables into a separate list for each optimizer
        grad_var_lists = [[] for _ in range(len(self.pred_opt_pairs) + 1)]
        for grad, var in grads_and_vars:
            if var.name in self.var_opt_mapping:
                grad_var_lists[self.var_opt_mapping[var.name]].append((grad, var))

        with tf.init_scope():
            for optimizer, opt_grads_and_vars in zip(self.optimizers, grad_var_lists):
                optimizer._create_slots([v for (_, v) in grads_and_vars])

        return tf.distribute.get_replica_context().merge_call(
            self._apply_gradients, args=(grad_var_lists, name), kwargs=kwargs
        )

    def _apply_gradients(self, distribution, grad_var_lists, name, **kwargs):
        # Apply gradients to each optimizer
        with tf.name_scope(self._name):
            train_ops = [
                distribution.extended.call_for_each_replica(
                    optimizer.apply_gradients, args=(opt_grads_and_vars,), kwargs=kwargs
                )
                for optimizer, opt_grads_and_vars in zip(
                    self.optimizers, grad_var_lists
                )
            ]

            return tf.group(*train_ops, name=name or "train_with_group")

    def get_config(self):
        optimizer_configs = [opt.get_config() for (_, opt) in self.pred_opt_pairs]
        default_config = self.default.get_config()

        config = {
            "optimizer_configs": [
                {"class_name": optimizer_config["name"], "config": optimizer_config}
                for optimizer_config in optimizer_configs
            ],
            "default_config": {
                "class_name": default_config["name"],
                "config": default_config,
            },
            "var_opt_mapping": self.var_opt_mapping,  # serialized instead of `pred`s
        }
        return {**super().get_config(), **config}

    @classmethod
    def from_config(cls, original_config, custom_objects=None):
        config = deepcopy(original_config)

        case_optimizer = cls(
            *[  # `(pred, opt)` tuples
                (
                    lambda _: False,  # placeholder callable (`pred` is not serialized)
                    tf.keras.optimizers.deserialize(  # optimizer `opt`
                        opt_config, custom_objects=custom_objects
                    ),
                )
                for opt_config in config["optimizer_configs"]
            ],
            default_optimizer=tf.keras.optimizers.deserialize(
                config["default_config"], custom_objects=custom_objects
            ),
        )

        # Since we no longer have the `pred`s, we set the mapping explicitly
        case_optimizer.var_opt_mapping = config["var_opt_mapping"]

        return case_optimizer

    def _compute_var_opt_mapping(self, grads_and_vars):
        """Compute a unique mapping from variables to optimizer indices."""

        self.var_opt_mapping = {}

        for grad, var in grads_and_vars:
            num_optimizers = 0

            # Find the optimizer(s) that want to claim this variable
            for optimizer_index, (predicate, _) in enumerate(self.pred_opt_pairs):
                if predicate(var):
                    self.var_opt_mapping[var.name] = optimizer_index
                    num_optimizers += 1

            if num_optimizers > 1:
                raise ValueError(f"Variable `{var}` claimed by multiple optimizers.")
            if num_optimizers == 0:
                if self.default is not None:
                    self.var_opt_mapping[var.name] = self.DEFAULT_OPT_INDEX
                else:
                    warnings.warn(
                        f"No `default_optimizer` provided to train variable `{var}`."
                    )

        # Make sure that each optimizer touches at least one variable
        for optimizer_index, (_, optimizer) in enumerate(self.pred_opt_pairs):
            if optimizer_index not in self.var_opt_mapping.values():
                raise ValueError(
                    f"Optimizer `{optimizer}` did not claim any variables."
                )


@utils.register_keras_custom_object
class Bop(tf.keras.optimizers.Optimizer):
    """Binary optimizer (Bop).

    Bop is a latent-free optimizer for Binarized Neural Networks (BNNs) and
    Binary Weight Networks (BWN).

    Bop maintains an exponential moving average of the gradients controlled by
    `gamma`. If this average exceeds the `threshold`, a weight is flipped.

    The hyperparameter `gamma` is somewhat analogues to the learning rate in
    SGD methods: a high `gamma` results in rapid convergence but also makes
    training more noisy.

    Note that the default `threshold` is not optimal for all situations.
    Setting the threshold too high results in little learning, while setting it
    too low results in overly noisy behaviour.

    !!! warning
        The `is_binary_variable` check of this optimizer will only target variables that
        have been explicitly marked as being binary using `NoOp(precision=1)`.

    !!! example
        ```python
        no_op_quantizer = lq.quantizers.NoOp(precision=1)
        layer = lq.layers.QuantDense(16, kernel_quantizer=no_op_quantizer)

        optimizer = lq.optimizers.CaseOptimizer(
            (lq.optimizers.Bop.is_binary_variable, lq.optimizers.Bop()),
            default_optimizer=tf.keras.optimizers.Adam(0.01),  # for FP weights
        )
        ```

    # Arguments
        threshold: magnitude of average gradient signal required to flip a weight.
        gamma: the adaptivity rate.
        name: name of the optimizer.

    # References
        - [Latent Weights Do Not Exist: Rethinking Binarized Neural Network Optimization](https://papers.nips.cc/paper/8971-latent-weights-do-not-exist-rethinking-binarized-neural-network-optimization)
    """

    _HAS_AGGREGATE_GRAD = True

    def __init__(
        self, threshold: float = 1e-8, gamma: float = 1e-4, name: str = "Bop", **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self._set_hyper("threshold", threshold)
        self._set_hyper("gamma", gamma)

    def _create_slots(self, var_list):
        for var in var_list:
            self.add_slot(var, "m")

    def _get_decayed_hyper(self, name: str, var_dtype):
        hyper = self._get_hyper(name, var_dtype)
        if isinstance(hyper, tf.keras.optimizers.schedules.LearningRateSchedule):
            local_step = tf.cast(self.iterations, var_dtype)
            hyper = tf.cast(hyper(local_step), var_dtype)
        return hyper

    def _resource_apply_dense(self, grad, var):
        var_dtype = var.dtype.base_dtype
        gamma = self._get_decayed_hyper("gamma", var_dtype)
        threshold = self._get_decayed_hyper("threshold", var_dtype)
        m = self.get_slot(var, "m")

        m_t = m.assign_add(gamma * (grad - m))
        var_t = lq.math.sign(-tf.sign(var * m_t - threshold) * var)
        return var.assign(var_t).op

    def get_config(self):
        config = {
            "threshold": self._serialize_hyperparameter("threshold"),
            "gamma": self._serialize_hyperparameter("gamma"),
        }
        return {**super().get_config(), **config}

    @classmethod
    def from_config(cls, config, custom_objects=None):
        for hyper in ("gamma", "threshold"):
            if hyper in config and isinstance(config[hyper], dict):
                config[hyper] = tf.keras.optimizers.schedules.deserialize(
                    config[hyper], custom_objects=custom_objects
                )
        return cls(**config)

    @staticmethod
    def is_binary_variable(var: tf.Variable) -> bool:
        """Returns `True` for variables with `var.precision == 1`.

        This is an example of a predictate that can be used by the `CaseOptimizer`.

        # Arguments
            var: a `tf.Variable`.
        """
        return getattr(var, "precision", 32) == 1



import numpy as np
import pytest
import tensorflow as tf
from packaging import version

import larq as lq
from larq import testing_utils


class DummyTrainableQuantizer(tf.keras.layers.Layer):
    """Used to test whether we can set layers as quantizers without any throws."""

    _custom_metrics = None

    def build(self, input_shape):
        self.dummy_weight = self.add_weight("dummy_weight", trainable=True)
        super().build(input_shape)

    def call(self, inputs):
        return self.dummy_weight * inputs


class TestCommonFunctionality:
    """Test functionality common to all quantizers, like serialization and usage."""

    @pytest.mark.parametrize("module", [lq.quantizers, tf.keras.activations])
    @pytest.mark.parametrize(
        "name,ref_cls",
        [
            ("ste_sign", lq.quantizers.SteSign),
            ("approx_sign", lq.quantizers.ApproxSign),
            ("ste_heaviside", lq.quantizers.SteHeaviside),
            ("magnitude_aware_sign", lq.quantizers.MagnitudeAwareSign),
            ("swish_sign", lq.quantizers.SwishSign),
            ("ste_tern", lq.quantizers.SteTern),
        ],
    )
    def test_serialization(self, module, name, ref_cls):
        fn = module.get(name)
        assert fn.__class__ == ref_cls
        fn = module.get(ref_cls())
        assert fn.__class__ == ref_cls
        assert type(fn.precision) == int
        if module == tf.keras.activations and version.parse(
            tf.__version__
        ) < version.parse("1.15"):
            pytest.skip(
                "TensorFlow < 1.15 does not support Quantizer classes as activations"
            )
        config = module.serialize(fn)
        fn = module.deserialize(config)
        assert fn.__class__ == ref_cls
        assert type(fn.precision) == int

    def test_noop_serialization(self):
        fn = lq.quantizers.get(lq.quantizers.NoOp(precision=1))
        assert fn.__class__ == lq.quantizers.NoOp
        assert fn.precision == 1
        config = lq.quantizers.serialize(fn)
        fn = lq.quantizers.deserialize(config)
        assert fn.__class__ == lq.quantizers.NoOp
        assert fn.precision == 1

    def test_invalid_usage(self):
        with pytest.raises(ValueError):
            lq.quantizers.get(42)
        with pytest.raises(ValueError):
            lq.quantizers.get("unknown")

    @pytest.mark.parametrize("quantizer", ["input_quantizer", "kernel_quantizer"])
    def test_layer_as_quantizer(self, quantizer, keras_should_run_eagerly):
        """Test whether a keras.layers.Layer can be used as quantizer."""

        input_data = testing_utils.random_input((1, 10))

        model = tf.keras.Sequential(
            [lq.layers.QuantDense(1, **{quantizer: DummyTrainableQuantizer()})]
        )
        model.compile(optimizer="sgd", loss="mse", run_eagerly=keras_should_run_eagerly)
        model.fit(input_data, np.ones((1,)), epochs=1)

        assert any(["dummy_weight" in var.name for var in model.trainable_variables])


class TestQuantization:
    """Test binarization and ternarization."""

    @pytest.mark.parametrize(
        "fn",
        [
            "ste_sign",
            lq.quantizers.SteSign(),
            "approx_sign",
            lq.quantizers.ApproxSign(),
            "swish_sign",
            lq.quantizers.SwishSign(),
        ],
    )
    def test_xnor_binarization(self, fn):
        x = tf.keras.backend.placeholder(ndim=2)
        f = tf.keras.backend.function([x], [lq.quantizers.get(fn)(x)])
        binarized_values = np.random.choice([-1, 1], size=(2, 5))
        result = f([binarized_values])[0]
        np.testing.assert_allclose(result, binarized_values)

        real_values = testing_utils.generate_real_values_with_zeros()
        result = f([real_values])[0]
        assert not np.any(result == 0)
        assert np.all(result[real_values < 0] == -1)
        assert np.all(result[real_values >= 0] == 1)

        zero_values = np.zeros((2, 5))
        result = f([zero_values])[0]
        assert np.all(result == 1)

    @pytest.mark.parametrize("fn", ["ste_heaviside", lq.quantizers.SteHeaviside()])
    def test_and_binarization(self, fn):
        x = tf.keras.backend.placeholder(ndim=2)
        f = tf.keras.backend.function([x], [lq.quantizers.get(fn)(x)])

        binarized_values = np.random.choice([0, 1], size=(2, 5))
        result = f([binarized_values])[0]
        np.testing.assert_allclose(result, binarized_values)

        real_values = testing_utils.generate_real_values_with_zeros()
        result = f([real_values])[0]
        assert np.all(result[real_values <= 0] == 0)
        assert np.all(result[real_values > 0] == 1)

    @pytest.mark.usefixtures("eager_mode")
    def test_magnitude_aware_sign_binarization(self):
        a = np.random.uniform(-2, 2, (3, 2, 2, 3))
        x = tf.Variable(a)
        y = lq.quantizers.MagnitudeAwareSign()(x)

        assert y.shape == x.shape

        # check sign
        np.testing.assert_allclose(tf.sign(y).numpy(), np.sign(a))

        # check magnitude
        np.testing.assert_allclose(
            tf.reduce_mean(tf.abs(y), axis=[0, 1, 2]).numpy(),
            [np.mean(np.reshape(np.abs(a[:, :, :, i]), [-1])) for i in range(3)],
        )

    @pytest.mark.parametrize(
        "fn",
        [
            "ste_tern",
            lq.quantizers.SteTern(),
            lq.quantizers.SteTern(ternary_weight_networks=True),
            lq.quantizers.SteTern(threshold_value=np.random.uniform(0.01, 0.8)),
        ],
    )
    def test_ternarization_basic(self, fn):
        x = tf.keras.backend.placeholder(ndim=2)
        f = tf.keras.backend.function([x], [lq.quantizers.get(fn)(x)])

        ternarized_values = np.random.choice([-1, 0, 1], size=(4, 10))
        result = f([ternarized_values])[0]
        np.testing.assert_allclose(result, ternarized_values)
        assert not np.any(result > 1)
        assert not np.any(result < -1)
        assert np.any(result == -1)
        assert np.any(result == 1)
        assert np.any(result == 0)

        real_values = testing_utils.generate_real_values_with_zeros()
        result = f([real_values])[0]
        assert not np.any(result > 1)
        assert not np.any(result < -1)
        assert np.any(result == -1)
        assert np.any(result == 1)
        assert np.any(result == 0)

    @pytest.mark.parametrize("fn", ["ste_tern", lq.quantizers.SteTern()])
    def test_ternarization_with_default_threshold(self, fn):
        x = tf.keras.backend.placeholder(ndim=2)
        test_threshold = 0.05  # This is the default
        f = tf.keras.backend.function([x], [lq.quantizers.get(fn)(x)])

        real_values = testing_utils.generate_real_values_with_zeros()
        result = f([real_values])[0]
        assert np.all(result[real_values > test_threshold] == 1)
        assert np.all(result[real_values < -test_threshold] == -1)
        assert np.all(result[np.abs(real_values) < test_threshold] == 0)
        assert not np.any(result > 1)
        assert not np.any(result < -1)

    def test_ternarization_with_custom_threshold(self):
        x = tf.keras.backend.placeholder(ndim=2)
        test_threshold = np.random.uniform(0.01, 0.8)
        fn = lq.quantizers.SteTern(threshold_value=test_threshold)
        f = tf.keras.backend.function([x], [fn(x)])

        real_values = testing_utils.generate_real_values_with_zeros()
        result = f([real_values])[0]
        assert np.all(result[real_values > test_threshold] == 1)
        assert np.all(result[real_values < -test_threshold] == -1)
        assert np.all(result[np.abs(real_values) < test_threshold] == 0)
        assert not np.any(result > 1)
        assert not np.any(result < -1)

    def test_ternarization_with_ternary_weight_networks(self):
        x = tf.keras.backend.placeholder(ndim=2)
        real_values = testing_utils.generate_real_values_with_zeros()
        test_threshold = 0.7 * np.sum(np.abs(real_values)) / np.size(real_values)
        fn = lq.quantizers.SteTern(ternary_weight_networks=True)
        f = tf.keras.backend.function([x], [fn(x)])

        result = f([real_values])[0]
        assert np.all(result[real_values > test_threshold] == 1)
        assert np.all(result[real_values < -test_threshold] == -1)
        assert np.all(result[np.abs(real_values) < test_threshold] == 0)
        assert not np.any(result > 1)
        assert not np.any(result < -1)

    def test_dorefa_quantize(self):
        x = tf.keras.backend.placeholder(ndim=2)
        f = tf.keras.backend.function([x], [lq.quantizers.DoReFa(2)(x)])
        real_values = testing_utils.generate_real_values_with_zeros()
        result = f([real_values])[0]
        k_bit = 2
        n = 2 ** k_bit - 1
        assert not np.any(result > 1)
        assert not np.any(result < 0)
        for i in range(n + 1):
            assert np.all(
                result[
                    (real_values > (2 * i - 1) / (2 * n))
                    & (real_values < (2 * i + 1) / (2 * n))
                ]
                == i / n
            )


@pytest.mark.usefixtures("eager_mode")
class TestGradients:
    """Test gradients for different quantizers."""

    @pytest.mark.parametrize(
        "fn",
        [
            lq.quantizers.SteSign(clip_value=None),
            lq.quantizers.SteTern(clip_value=None),
            lq.quantizers.SteHeaviside(clip_value=None),
        ],
    )
    def test_identity_ste_grad(self, fn):
        x = testing_utils.generate_real_values_with_zeros(shape=(8, 3, 3, 16))
        tf_x = tf.Variable(x)
        with tf.GradientTape() as tape:
            activation = fn(tf_x)
        grad = tape.gradient(activation, tf_x)
        np.testing.assert_allclose(grad.numpy(), np.ones_like(x))

    @pytest.mark.parametrize(
        "fn",
        [
            lq.quantizers.SteSign(),
            lq.quantizers.SteTern(),
            lq.quantizers.SteHeaviside(),
        ],
    )
    def test_ste_grad(self, fn):
        @np.vectorize
        def ste_grad(x):
            if np.abs(x) <= 1:
                return 1.0
            return 0.0

        x = testing_utils.generate_real_values_with_zeros(shape=(8, 3, 3, 16))
        tf_x = tf.Variable(x)
        with tf.GradientTape() as tape:
            activation = fn(tf_x)
        grad = tape.gradient(activation, tf_x)
        np.testing.assert_allclose(grad.numpy(), ste_grad(x))

    # Test with and without default threshold
    def test_swish_grad(self):
        def swish_grad(x, beta):
            return (
                beta * (2 - beta * x * np.tanh(beta * x / 2)) / (1 + np.cosh(beta * x))
            )

        x = testing_utils.generate_real_values_with_zeros(shape=(8, 3, 3, 16))
        tf_x = tf.Variable(x)
        with tf.GradientTape() as tape:
            activation = lq.quantizers.SwishSign()(tf_x)
        grad = tape.gradient(activation, tf_x)
        np.testing.assert_allclose(grad.numpy(), swish_grad(x, beta=5.0))

        with tf.GradientTape() as tape:
            activation = lq.quantizers.SwishSign(beta=10.0)(tf_x)
        grad = tape.gradient(activation, tf_x)
        np.testing.assert_allclose(grad.numpy(), swish_grad(x, beta=10.0))

    def test_approx_sign_grad(self):
        @np.vectorize
        def approx_sign_grad(x):
            if np.abs(x) <= 1:
                return 2 - 2 * np.abs(x)
            return 0.0

        x = testing_utils.generate_real_values_with_zeros(shape=(8, 3, 3, 16))
        tf_x = tf.Variable(x)
        with tf.GradientTape() as tape:
            activation = lq.quantizers.ApproxSign()(tf_x)
        grad = tape.gradient(activation, tf_x)
        np.testing.assert_allclose(grad.numpy(), approx_sign_grad(x))

    def test_magnitude_aware_sign_grad(self):
        a = np.random.uniform(-2, 2, (3, 2, 2, 3))
        x = tf.Variable(a)
        with tf.GradientTape() as tape:
            y = lq.quantizers.MagnitudeAwareSign()(x)
        grad = tape.gradient(y, x)

        scale_vector = [
            np.mean(np.reshape(np.abs(a[:, :, :, i]), [-1])) for i in range(3)
        ]

        np.testing.assert_allclose(
            grad.numpy(), np.where(abs(a) < 1, np.ones(a.shape) * scale_vector, 0)
        )

    def test_dorefa_ste_grad(self):
        @np.vectorize
        def ste_grad(x):
            if x <= 1 and x >= 0:
                return 1.0
            return 0.0

        x = testing_utils.generate_real_values_with_zeros(shape=(8, 3, 3, 16))
        tf_x = tf.Variable(x)
        with tf.GradientTape() as tape:
            activation = lq.quantizers.DoReFa(2)(tf_x)
        grad = tape.gradient(activation, tf_x)
        np.testing.assert_allclose(grad.numpy(), ste_grad(x))


@pytest.mark.parametrize(
    "quantizer",
    [
        ("ste_sign", lq.quantizers.SteSign),
        ("approx_sign", lq.quantizers.ApproxSign),
        ("ste_heaviside", lq.quantizers.SteHeaviside),
        ("swish_sign", lq.quantizers.SwishSign),
        ("magnitude_aware_sign", lq.quantizers.MagnitudeAwareSign),
        ("ste_tern", lq.quantizers.SteTern),
        ("dorefa_quantizer", lq.quantizers.DoReFa),
    ],
)
def test_metrics(quantizer):
    quantizer_str, quantizer_cls = quantizer

    # No metric
    model = tf.keras.models.Sequential(
        [lq.layers.QuantDense(3, kernel_quantizer=quantizer_str, input_shape=(32,))]
    )
    model.compile(loss="mse", optimizer="sgd")
    assert len(model.layers[0]._metrics) == 0

    # Metric added using scope
    with lq.context.metrics_scope(["flip_ratio"]):
        model = tf.keras.models.Sequential(
            [lq.layers.QuantDense(3, kernel_quantizer=quantizer_str, input_shape=(32,))]
        )
    model.compile(loss="mse", optimizer="sgd")

    if version.parse(tf.__version__) > version.parse("1.14"):
        assert len(model.layers[0].kernel_quantizer._metrics) == 1
    else:
        # In TF1.14, call() gets called twice, resulting in having an extra initial
        # metrics copy.
        assert len(model.layers[0].kernel_quantizer._metrics) == 2

    # Metric added explicitly to quantizer
    model = tf.keras.models.Sequential(
        [
            lq.layers.QuantDense(
                3,
                kernel_quantizer=quantizer_cls(metrics=["flip_ratio"]),
                input_shape=(32,),
            )
        ]
    )
    model.compile(loss="mse", optimizer="sgd")
    if version.parse(tf.__version__) > version.parse("1.14"):
        assert len(model.layers[0].kernel_quantizer._metrics) == 1
    else:
        # In TF1.14, call() gets called twice, resulting in having an extra initial
        # metrics copy.
        assert len(model.layers[0].kernel_quantizer._metrics) == 2


def test_get_kernel_quantizer_assigns_metrics():
    with lq.context.metrics_scope(["flip_ratio"]):
        ste_sign = lq.quantizers.get_kernel_quantizer("ste_sign")
        assert "flip_ratio" in lq.context.get_training_metrics()

    assert isinstance(ste_sign, lq.quantizers.SteSign)
    assert "flip_ratio" in ste_sign._custom_metrics


def test_get_kernel_quantizer_accepts_function():
    custom_quantizer = lq.quantizers.get_kernel_quantizer(lambda x: x)
    assert callable(custom_quantizer)
    assert not hasattr(custom_quantizer, "_custom_metrics")


def test_backwards_compat_aliases():
    assert lq.quantizers.DoReFaQuantizer == lq.quantizers.DoReFa
    assert lq.quantizers.NoOpQuantizer == lq.quantizers.NoOp



import numpy as np
import pytest
import tensorflow as tf
from tensorflow import keras
from tensorflow.python.keras import testing_utils

import larq as lq
from larq import testing_utils as lq_testing_utils


def _assert_weights(weights, expected):
    for w, e in zip(weights, expected):
        np.testing.assert_allclose(np.squeeze(w), e)


def _test_optimizer(
    optimizer, target=0.75, test_kernels_are_binary=True, trainable_bn=True
):
    np.random.seed(1337)
    (x_train, y_train), _ = testing_utils.get_test_data(
        train_samples=1000, test_samples=0, input_shape=(10,), num_classes=2
    )
    y_train = keras.utils.to_categorical(y_train)

    model = lq_testing_utils.get_small_bnn_model(
        x_train.shape[1], 20, y_train.shape[1], trainable_bn=trainable_bn
    )
    model.compile(loss="categorical_crossentropy", optimizer=optimizer, metrics=["acc"])

    initial_vars = [tf.keras.backend.get_value(w) for w in model.trainable_weights]

    history = model.fit(x_train, y_train, epochs=2, batch_size=16, verbose=0)

    trained_vars = [tf.keras.backend.get_value(w) for w in model.trainable_weights]

    # check all trainable variables have actually been updated
    for v0, v1 in zip(initial_vars, trained_vars):
        assert not np.all(v0 == v1)

    # Note that when kernels are treated as latent weights they need not be
    # binary (see https://arxiv.org/abs/1906.02107 for further discussion)
    if test_kernels_are_binary:
        for layer in model.layers:
            if "quant" in layer.name:
                for weight in layer.trainable_weights:
                    assert np.all(np.isin(tf.keras.backend.get_value(weight), [-1, 1]))

    assert history.history["acc"][-1] >= target


def _test_serialization(optimizer):
    config = keras.optimizers.serialize(optimizer)
    optim = keras.optimizers.deserialize(config)
    new_config = keras.optimizers.serialize(optim)
    assert config == new_config


class TestCaseOptimizer:
    def test_type_check_predicate(self):
        with pytest.raises(TypeError):
            lq.optimizers.CaseOptimizer((False, lq.optimizers.Bop()))

    def test_type_check_optimizer(self):
        with pytest.raises(TypeError):
            lq.optimizers.CaseOptimizer((lq.optimizers.Bop.is_binary_variable, False))

    def test_type_check_default(self):
        with pytest.raises(TypeError):
            lq.optimizers.CaseOptimizer(
                (lq.optimizers.Bop.is_binary_variable, lq.optimizers.Bop()),
                default_optimizer=False,
            )

    def test_overlapping_predicates(self):
        with pytest.raises(ValueError):
            naughty_case_opt = lq.optimizers.CaseOptimizer(
                (lambda var: True, lq.optimizers.Bop()),
                (lambda var: True, lq.optimizers.Bop()),
            )
            _test_optimizer(naughty_case_opt)

    def test_missing_default(self):
        with pytest.warns(Warning):
            naughty_case_opt = lq.optimizers.CaseOptimizer(
                (lq.optimizers.Bop.is_binary_variable, lq.optimizers.Bop()),
            )

            # Simple MNIST model
            mnist = tf.keras.datasets.mnist
            (train_images, train_labels), _ = mnist.load_data()
            model = tf.keras.Sequential(
                [
                    tf.keras.layers.Flatten(input_shape=(28, 28)),
                    lq.layers.QuantDense(
                        64,
                        input_quantizer="ste_sign",
                        kernel_quantizer=lq.quantizers.NoOp(precision=1),
                        activation="relu",
                    ),
                    tf.keras.layers.Dense(10, activation="softmax"),
                ]
            )
            model.compile(
                loss="sparse_categorical_crossentropy",
                optimizer=naughty_case_opt,
                metrics=["acc"],
            )

            # Should raise on first call to apply_gradients()
            model.fit(train_images[:1], train_labels[:1], epochs=1)

    def test_wrong_predicate(self):
        """Make sure we throw when an optimizer does not claim variables."""

        with pytest.raises(ValueError):
            naughty_case_opt = lq.optimizers.CaseOptimizer(
                (lambda var: False, lq.optimizers.Bop()),
                default_optimizer=tf.keras.optimizers.Adam(0.01),
            )

            # Simple MNIST model
            mnist = tf.keras.datasets.mnist
            (train_images, train_labels), _ = mnist.load_data()
            model = tf.keras.Sequential(
                [
                    tf.keras.layers.Flatten(input_shape=(28, 28)),
                    tf.keras.layers.Dense(128, activation="relu"),
                    tf.keras.layers.Dense(10, activation="softmax"),
                ]
            )
            model.compile(
                loss="sparse_categorical_crossentropy",
                optimizer=naughty_case_opt,
                metrics=["acc"],
            )

            # Should raise on first call to apply_gradients()
            model.fit(train_images[:1], train_labels[:1], epochs=1)

    def test_weights(self):
        (train_images, train_labels), _ = tf.keras.datasets.mnist.load_data()
        model = tf.keras.Sequential(
            [
                tf.keras.layers.Flatten(input_shape=(28, 28)),
                lq.layers.QuantDense(
                    64,
                    input_quantizer="ste_sign",
                    kernel_quantizer=lq.quantizers.NoOp(precision=1),
                    activation="relu",
                ),
                tf.keras.layers.Dense(10, activation="softmax"),
            ]
        )
        model.compile(
            loss="sparse_categorical_crossentropy",
            optimizer=lq.optimizers.CaseOptimizer(
                (lq.optimizers.Bop.is_binary_variable, lq.optimizers.Bop()),
                default_optimizer=tf.keras.optimizers.Adam(0.01),
            ),
        )
        model.fit(train_images[:1], train_labels[:1], epochs=1)

        opt_weights = model.optimizer.weights
        assert len(opt_weights) != 0
        checked_weights = 0
        for opt in model.optimizer.optimizers:
            for weight in opt.weights:
                assert weight is opt_weights[checked_weights]
                checked_weights += 1
        assert checked_weights == len(opt_weights)

    @pytest.mark.usefixtures("eager_mode")
    def test_checkpoint(self, tmp_path):
        # Build and run a simple model.
        var = tf.Variable([2.0])
        opt = tf.keras.optimizers.SGD(1.0, momentum=1.0)
        opt = lq.optimizers.CaseOptimizer((lambda var: True, opt))
        opt.minimize(lambda: var + 1.0, var_list=[var])
        slot_var = opt.optimizers[0].get_slot(var, "momentum")
        slot_value = slot_var.numpy().item()

        # Save a checkpoint.
        checkpoint = tf.train.Checkpoint(optimizer=opt, var=var)
        save_path = checkpoint.save(tmp_path / "ckpt")

        # Run model again.
        opt.minimize(lambda: var + 1.0, var_list=[var])
        assert slot_var.numpy().item() != slot_value

        # Load checkpoint and ensure loss scale is back to its original value.
        status = checkpoint.restore(save_path)
        status.assert_consumed()
        status.run_restore_ops()
        assert slot_var.numpy().item() == slot_value


class TestBopOptimizer:
    def test_bop_accuracy(self):
        _test_optimizer(
            lq.optimizers.CaseOptimizer(
                (lq.optimizers.Bop.is_binary_variable, lq.optimizers.Bop()),
                default_optimizer=tf.keras.optimizers.Adam(0.01),
            ),
            test_kernels_are_binary=True,
        )
        # test optimizer on model with only binary trainable vars (low accuracy)
        _test_optimizer(
            lq.optimizers.CaseOptimizer(
                (lq.optimizers.Bop.is_binary_variable, lq.optimizers.Bop()),
                default_optimizer=tf.keras.optimizers.Adam(0.01),
            ),
            test_kernels_are_binary=True,
            trainable_bn=False,
            target=0,
        )

    @pytest.mark.usefixtures("distribute_scope")
    def test_mixed_precision(self):
        opt = lq.optimizers.CaseOptimizer(
            (lq.optimizers.Bop.is_binary_variable, lq.optimizers.Bop()),
            default_optimizer=tf.keras.optimizers.Adam(0.01),
        )
        opt = tf.keras.mixed_precision.experimental.LossScaleOptimizer(opt, "dynamic")
        _test_optimizer(opt, test_kernels_are_binary=True)

    def test_bop_tf_1_14_schedules(self):
        _test_optimizer(
            lq.optimizers.CaseOptimizer(
                (
                    lq.optimizers.Bop.is_binary_variable,
                    lq.optimizers.Bop(
                        threshold=tf.keras.optimizers.schedules.InverseTimeDecay(
                            3.0, decay_steps=1.0, decay_rate=0.5
                        ),
                        gamma=tf.keras.optimizers.schedules.InverseTimeDecay(
                            3.0, decay_steps=1.0, decay_rate=0.5
                        ),
                    ),
                ),
                default_optimizer=tf.keras.optimizers.Adam(0.01),
            ),
            test_kernels_are_binary=True,
        )

    def test_bop_serialization(self):
        _test_serialization(
            lq.optimizers.CaseOptimizer(
                (lq.optimizers.Bop.is_binary_variable, lq.optimizers.Bop()),
                default_optimizer=tf.keras.optimizers.Adam(0.01),
            ),
        )

    @pytest.mark.parametrize(
        "hyper",
        [5e-4, tf.keras.optimizers.schedules.PolynomialDecay(5e-4, 100)],
    )
    def test_bop_serialization_schedule(self, hyper):
        bop = lq.optimizers.Bop(
            gamma=hyper,
            threshold=hyper,
        )
        new_bop = lq.optimizers.Bop.from_config(bop.get_config())
        assert isinstance(new_bop._get_hyper("gamma"), type(bop._get_hyper("gamma")))
        assert isinstance(
            new_bop._get_hyper("threshold"), type(bop._get_hyper("threshold"))
        )



import numpy as np
import pytest
import tensorflow as tf

import larq as lq
from larq.testing_utils import generate_real_values_with_zeros


@pytest.mark.parametrize("name", ["hard_tanh", "leaky_tanh"])
def test_serialization(name):
    fn = tf.keras.activations.get(name)
    ref_fn = getattr(lq.activations, name)
    assert fn == ref_fn
    config = tf.keras.activations.serialize(fn)
    fn = tf.keras.activations.deserialize(config)
    assert fn == ref_fn


def test_hard_tanh():
    real_values = generate_real_values_with_zeros()
    x = tf.keras.backend.placeholder(ndim=2)
    f = tf.keras.backend.function([x], [lq.activations.hard_tanh(x)])
    result = f([real_values])[0]
    np.testing.assert_allclose(result, np.clip(real_values, -1, 1))


def test_leaky_tanh():
    @np.vectorize
    def leaky_tanh(x, alpha):
        if x <= -1:
            return -1 + alpha * (x + 1)
        elif x <= 1:
            return x
        else:
            return 1 + alpha * (x - 1)

    real_values = generate_real_values_with_zeros()
    x = tf.keras.backend.placeholder(ndim=2)
    f = tf.keras.backend.function([x], [lq.activations.leaky_tanh(x)])
    result = f([real_values])[0]
    np.testing.assert_allclose(result, leaky_tanh(real_values, alpha=0.2))



import numpy as np
import pytest
import tensorflow as tf
from packaging import version

import larq as lq
from larq.models import ModelProfile


class ToyModel(tf.keras.Model):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.conv = lq.layers.QuantConv2D(
            filters=32,
            kernel_size=(3, 3),
            kernel_quantizer="ste_sign",
            input_shape=(64, 64, 1),
            padding="same",
        )
        self.pool = tf.keras.layers.GlobalAvgPool2D()
        self.dense = tf.keras.layers.Dense(10, activation="softmax")

    def call(self, inputs):
        return self.dense(self.pool(self.conv(inputs)))


def get_functional_model():
    input = tf.keras.Input((32, 32, 3))
    x = lq.layers.QuantConv2D(
        filters=32,
        kernel_size=(3, 3),
        kernel_quantizer="ste_sign",
        padding="same",
    )(input)
    y, z = tf.split(x, 2, axis=-1)
    x = tf.concat([y, z], axis=-1)
    return tf.keras.Model(input, x, name="toy_model")


def get_profile_model():
    return tf.keras.models.Sequential(
        [
            lq.layers.QuantConv2D(
                filters=32,
                kernel_size=(3, 3),
                kernel_quantizer="ste_sign",
                input_shape=(64, 64, 1),
                padding="same",
            ),
            tf.keras.layers.MaxPooling2D((2, 2)),
            lq.layers.QuantDepthwiseConv2D(
                kernel_size=3,
                strides=(3, 3),
                input_quantizer=lq.quantizers.SteTern(),
                depthwise_quantizer=lq.quantizers.SteTern(),
                padding="same",
                pad_values=1.0,
                use_bias=False,
            ),
            tf.keras.layers.BatchNormalization(scale=False),
            lq.layers.QuantSeparableConv2D(
                32,
                (3, 3),
                input_quantizer="ste_sign",
                depthwise_quantizer="ste_sign",
                pointwise_quantizer="ste_sign",
                padding="same",
            ),
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(10, trainable=False),
        ]
    )


def test_model_profile():
    profile = ModelProfile(get_profile_model())
    assert len(profile.layer_profiles) == 7


def test_layer_profile():
    profile = ModelProfile(get_profile_model())

    kernel_count = [
        32 * 3 * 3 * 1,
        0,
        32 * 3 * 3,
        0,
        32 * 3 * 3 * 1 + 32 * 1 * 1 * 32,
        0,
        32 * 11 * 11 * 10,
    ]
    bias_count = [32, 0, 0, 64, 32, 0, 10]
    param_count = [k + b for k, b in zip(kernel_count, bias_count)]
    memory = [  # bits * (c * w * h * b) + bits * bias
        1 * (32 * 3 * 3 * 1) + 32 * 32,
        0,
        2 * (32 * 3 * 3),
        32 * (2 * 32),
        1 * (32 * 3 * 3 * 1 + 32 * 1 * 1 * 32) + 32 * 32,
        0,
        32 * (32 * 11 * 11 * 10 + 10),
    ]
    int8_fp_weights_mem = [
        1 * (32 * 3 * 3 * 1) + 8 * 32,
        0,
        2 * (32 * 3 * 3),
        8 * (32 * 2),
        1 * (32 * 3 * 3 * 1 + 32 * 1 * 1 * 32) + 8 * 32,
        0,
        8 * (32 * 11 * 11 * 10 + 10),
    ]
    fp_equiv_mem = [32 * n for n in param_count]
    input_precision = [None, None, 2, None, 1, None, None]
    output_shape = [
        (-1, 64, 64, 32),
        (-1, 32, 32, 32),
        (-1, 11, 11, 32),
        (-1, 11, 11, 32),
        (-1, 11, 11, 32),
        (-1, 11 * 11 * 32),
        (-1, 10),
    ]
    output_pixels = [int(np.prod(os[1:-1])) for os in output_shape]
    unique_param_bidtwidths = [[1, 32], [], [2], [32], [1, 32], [], [32]]
    unique_op_precisions = [[32], [], [2], [], [1], [], [32]]
    mac_count = [params * pixels for params, pixels in zip(kernel_count, output_pixels)]
    bin_mac_count = [
        mc if (1 in pb and ip == 1) else 0
        for mc, pb, ip in zip(mac_count, unique_param_bidtwidths, input_precision)
    ]

    profiles = profile.layer_profiles
    for i in range(len(profiles)):
        print(f"Testing layer {i}...")
        assert profiles[i].input_precision == input_precision[i]
        assert profiles[i].output_shape == output_shape[i]
        assert profiles[i].output_pixels == output_pixels[i]
        assert profiles[i].weight_count() == param_count[i]
        assert profiles[i].unique_param_bidtwidths == unique_param_bidtwidths[i]
        assert profiles[i].unique_op_precisions == unique_op_precisions[i]
        assert profiles[i].memory == memory[i]
        assert profiles[i].fp_equivalent_memory == fp_equiv_mem[i]
        assert profiles[i].int8_fp_weights_memory == int8_fp_weights_mem[i]
        assert profiles[i].op_count("mac") == mac_count[i]
        assert profiles[i].op_count("mac", 1) == bin_mac_count[i]


def test_layer_profile_1d():
    model = tf.keras.models.Sequential(
        [
            lq.layers.QuantConv1D(
                filters=32,
                kernel_size=3,
                input_shape=(64, 6),
                kernel_quantizer="ste_sign",
                padding="same",
            ),
            tf.keras.layers.MaxPooling1D(2),
            lq.layers.QuantSeparableConv1D(
                filters=16,
                kernel_size=3,
                input_quantizer="ste_sign",
                depthwise_quantizer="ste_sign",
                pointwise_quantizer="ste_sign",
                padding="same",
            ),
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(10, trainable=False),
        ]
    )
    profile = ModelProfile(model)

    kernel_count = [(32 * 3 * 6), 0, (32 * 3 + 16 * 32), 0, (16 * 32 * 10)]
    bias_count = [32, 0, 16, 0, 10]
    param_count = [k + b for k, b in zip(kernel_count, bias_count)]
    memory = [  # bits * (c * w * d) + bits * bias
        1 * (32 * 3 * 6) + 32 * 32,
        0,
        1 * (32 * 3 + 16 * 32) + 32 * 16,
        0,
        32 * (32 * 16 * 10 + 10),
    ]
    int8_fp_weights_mem = [
        1 * (32 * 3 * 6) + 8 * 32,
        0,
        1 * (32 * 3 + 16 * 32) + 8 * 16,
        0,
        8 * (32 * 16 * 10 + 10),
    ]
    fp_equiv_mem = [32 * n for n in param_count]
    input_precision = [None, None, 1, None, None]
    output_shape = [
        (-1, 64, 32),
        (-1, 32, 32),
        (-1, 32, 16),
        (-1, 32 * 16),
        (-1, 10),
    ]
    output_pixels = [int(np.prod(os[1:-1])) for os in output_shape]
    unique_param_bidtwidths = [[1, 32], [], [1, 32], [], [32]]
    unique_op_precisions = [[32], [], [1], [], [32]]
    mac_count = [params * pixels for params, pixels in zip(kernel_count, output_pixels)]
    bin_mac_count = [
        mc if (1 in pb and ip == 1) else 0
        for mc, pb, ip in zip(mac_count, unique_param_bidtwidths, input_precision)
    ]

    profiles = profile.layer_profiles
    for i in range(len(profiles)):
        print(f"Testing layer {i}...")
        assert profiles[i].input_precision == input_precision[i]
        assert profiles[i].output_shape == output_shape[i]
        assert profiles[i].output_pixels == output_pixels[i]
        assert profiles[i].weight_count() == param_count[i]
        assert profiles[i].unique_param_bidtwidths == unique_param_bidtwidths[i]
        assert profiles[i].unique_op_precisions == unique_op_precisions[i]
        assert profiles[i].memory == memory[i]
        assert profiles[i].fp_equivalent_memory == fp_equiv_mem[i]
        assert profiles[i].int8_fp_weights_memory == int8_fp_weights_mem[i]
        assert profiles[i].op_count("mac") == mac_count[i]
        assert profiles[i].op_count("mac", 1) == bin_mac_count[i]


def test_summary(snapshot, capsys):
    model = get_profile_model()
    lq.models.summary(model)
    captured = capsys.readouterr()
    snapshot.assert_match(captured.out)

    # A model with no weights
    model = tf.keras.models.Sequential(
        [tf.keras.layers.Lambda(lambda x: tf.zeros(2), input_shape=(32, 32))]
    )
    lq.models.summary(model)
    captured = capsys.readouterr()
    snapshot.assert_match(captured.out)


def test_subclass_model_summary(snapshot, capsys):
    model = ToyModel()
    model.build((None, 32, 32, 3))
    lq.models.summary(model)
    captured = capsys.readouterr()
    snapshot.assert_match(captured.out)


def test_functional_model_summary(snapshot, capsys):
    lq.models.summary(get_functional_model())
    captured = capsys.readouterr()
    key = "2.4+" if version.parse(tf.__version__) >= version.parse("2.3.9") else "<2.4"
    snapshot.assert_match(captured.out.lower(), key)


def test_summary_invalid_model():
    with pytest.raises(ValueError):
        lq.models.summary(tf.keras.Model())


def test_bitsize_invalid_key():
    with pytest.raises(NotImplementedError):
        lq.models._bitsize_as_str(-1)


def test_number_as_readable_str_large():
    assert lq.models._number_as_readable_str(1e16) == "1.00E+16"


@pytest.fixture(autouse=True)
def run_around_tests():
    tf.keras.backend.clear_session()
    yield



import numpy as np
import pytest
import tensorflow as tf
from numpy.testing import assert_almost_equal, assert_array_equal
from packaging import version
from tensorflow.python.distribute.values import DistributedVariable

from larq import context, testing_utils
from larq.quantized_variable import QuantizedVariable
from larq.testing_utils import evaluate


def get_var(val, dtype=None, name=None):
    return tf.compat.v1.Variable(val, use_resource=True, dtype=dtype, name=name)


def test_inheritance(distribute_scope):
    variable = get_var(3.0)
    quantized_variable = QuantizedVariable.from_variable(variable)
    assert isinstance(quantized_variable, QuantizedVariable)
    assert isinstance(quantized_variable, tf.Variable)
    assert isinstance(quantized_variable, DistributedVariable) is distribute_scope


@pytest.mark.usefixtures("eager_and_graph_mode", "distribute_scope")
def test_read():
    x = QuantizedVariable.from_variable(get_var(3.5), quantizer=lambda x: 2 * x)
    evaluate(x.initializer)

    assert evaluate(x) == 3.5
    assert evaluate(x.value()) == 3.5
    assert evaluate(x.read_value()) == 3.5
    assert evaluate(tf.identity(x)) == 3.5

    with context.quantized_scope(True):
        assert evaluate(x) == 7
        assert evaluate(x.value()) == 7
        assert evaluate(x.read_value()) == 7
        assert evaluate(tf.identity(x)) == 7


@pytest.mark.usefixtures("eager_and_graph_mode")
def test_sparse_reads():
    x = QuantizedVariable.from_variable(get_var([1.0, 2.0]), quantizer=lambda x: 2 * x)
    evaluate(x.initializer)

    assert evaluate(x.sparse_read([0])) == 1
    assert evaluate(x.gather_nd([0])) == 1
    with context.quantized_scope(True):
        assert evaluate(x.sparse_read([0])) == 2
        assert evaluate(x.gather_nd([0])) == 2


@pytest.mark.usefixtures("eager_and_graph_mode", "distribute_scope")
def test_read_nested_scopes():
    x = QuantizedVariable.from_variable(get_var(3.5), quantizer=lambda x: 2 * x)
    evaluate(x.initializer)
    with context.quantized_scope(True):
        assert evaluate(x.read_value()) == 7
        with context.quantized_scope(False):
            assert evaluate(x.read_value()) == 3.5
        assert evaluate(x.read_value()) == 7


@pytest.mark.usefixtures("eager_and_graph_mode")
def test_method_delegations(distribute_scope):
    x = QuantizedVariable.from_variable(get_var(3.5), quantizer=lambda x: 2 * x)
    with context.quantized_scope(True):
        evaluate(x.initializer)
        assert evaluate(x.value()) == 7
        assert evaluate(x.read_value()) == 7
        assert x.trainable
        if version.parse(tf.__version__) > version.parse("1.14"):
            assert x.synchronization == x.latent_variable.synchronization
        assert x.aggregation == x.latent_variable.aggregation
        assert evaluate(x.initialized_value()) == 7
        if not tf.executing_eagerly():
            if not distribute_scope:
                # These functions are not supported for DistributedVariables
                x.load(4.5)
                assert x.eval() == 9
            assert evaluate(x.initial_value) == 7
            assert x.op == x.latent_variable.op
            assert x.graph == x.latent_variable.graph
        if not distribute_scope:
            # These attributes are not supported for DistributedVariables
            assert x.constraint is None
            assert x.initializer == x.latent_variable.initializer

        def apply_and_read(x, fn, args):
            evaluate(fn(*args))
            return evaluate(x)

        assert apply_and_read(x, x.assign, [4]) == 8
        assert apply_and_read(x, x.assign_add, [1]) == 10
        assert apply_and_read(x, x.assign_sub, [1.5]) == 7
        assert x.name == x.latent_variable.name
        assert x.device == x.latent_variable.device
        assert x.shape == ()
        assert x.get_shape() == ()
        try:
            x.set_shape(())
            assert x.shape == ()
        except NotImplementedError:
            pass


@pytest.mark.usefixtures("eager_and_graph_mode")
def test_scatter_method_delegations():
    x = QuantizedVariable.from_variable(get_var([3.5, 4]), quantizer=lambda x: 2 * x)
    evaluate(x.initializer)
    with context.quantized_scope(True):
        assert_array_equal(evaluate(x.value()), [7, 8])

        def slices(val, index):
            return tf.IndexedSlices(values=val, indices=index)

        assert_array_equal(evaluate(x.scatter_sub(slices(0.5, 0))), [6, 8])
        assert_array_equal(evaluate(x.scatter_add(slices(0.5, 0))), [7, 8])
        if version.parse(tf.__version__) > version.parse("1.14"):
            assert_array_equal(evaluate(x.scatter_max(slices(4.5, 1))), [7, 9])
            assert_array_equal(evaluate(x.scatter_min(slices(4.0, 1))), [7, 8])
            assert_array_equal(evaluate(x.scatter_mul(slices(2.0, 1))), [7, 16])
            assert_array_equal(evaluate(x.scatter_div(slices(2.0, 1))), [7, 8])
        assert_array_equal(evaluate(x.scatter_update(slices(2.0, 1))), [7, 4])
        assert_array_equal(evaluate(x.scatter_nd_sub([[0], [1]], [0.5, 1.0])), [6, 2])
        assert_array_equal(evaluate(x.scatter_nd_add([[0], [1]], [0.5, 1.0])), [7, 4])
        assert_array_equal(
            evaluate(x.scatter_nd_update([[0], [1]], [0.5, 1.0])), [1, 2]
        )
        assert_array_equal(evaluate(x.batch_scatter_update(slices([2.0], [1]))), [1, 4])


@pytest.mark.usefixtures("eager_and_graph_mode", "distribute_scope")
def test_overloads(quantized):
    if quantized:
        x = QuantizedVariable.from_variable(get_var(3.5), quantizer=lambda x: 2 * x)
    else:
        x = QuantizedVariable.from_variable(get_var(7.0))
    evaluate(x.initializer)
    assert_almost_equal(8, evaluate(x + 1))
    assert_almost_equal(10, evaluate(3 + x))
    assert_almost_equal(14, evaluate(x + x))
    assert_almost_equal(5, evaluate(x - 2))
    assert_almost_equal(6, evaluate(13 - x))
    assert_almost_equal(0, evaluate(x - x))
    assert_almost_equal(14, evaluate(x * 2))
    assert_almost_equal(21, evaluate(3 * x))
    assert_almost_equal(49, evaluate(x * x))
    assert_almost_equal(3.5, evaluate(x / 2))
    assert_almost_equal(1.5, evaluate(10.5 / x))
    assert_almost_equal(3, evaluate(x // 2))
    assert_almost_equal(2, evaluate(15 // x))
    assert_almost_equal(1, evaluate(x % 2))
    assert_almost_equal(2, evaluate(16 % x))
    assert evaluate(x < 12)
    assert evaluate(x <= 12)
    assert not evaluate(x > 12)
    assert not evaluate(x >= 12)
    assert not evaluate(12 < x)
    assert not evaluate(12 <= x)
    assert evaluate(12 > x)
    assert evaluate(12 >= x)
    assert_almost_equal(343, evaluate(pow(x, 3)))
    assert_almost_equal(128, evaluate(pow(2, x)))
    assert_almost_equal(-7, evaluate(-x))
    assert_almost_equal(7, evaluate(abs(x)))


@pytest.mark.usefixtures("eager_mode")
def test_tensor_equality(quantized):
    if quantized:
        x = QuantizedVariable.from_variable(
            get_var([3.5, 4.0, 4.5]), quantizer=lambda x: 2 * x
        )
    else:
        x = QuantizedVariable.from_variable(get_var([7.0, 8.0, 9.0]))
    evaluate(x.initializer)
    assert_array_equal(evaluate(x), [7.0, 8.0, 9.0])
    if version.parse(tf.__version__) >= version.parse("2"):
        assert_array_equal(x == [7.0, 8.0, 10.0], [True, True, False])
        assert_array_equal(x != [7.0, 8.0, 10.0], [False, False, True])


@pytest.mark.usefixtures("eager_and_graph_mode")
def test_assign(quantized, distribute_scope):
    x = QuantizedVariable.from_variable(
        get_var(0.0, tf.float64), quantizer=lambda x: 2 * x
    )
    evaluate(x.initializer)

    latent_value = 3.14
    value = latent_value * 2 if quantized else latent_value

    # Assign doesn't correctly return a quantized variable in graph mode if a strategy is used
    if tf.executing_eagerly() or not distribute_scope or not quantized:
        # Assign float32 values
        lv = tf.constant(latent_value, dtype=tf.float64)
        assert_almost_equal(evaluate(x.assign(lv)), value)
        assert_almost_equal(evaluate(x.assign_add(lv)), value * 2)
        assert_almost_equal(evaluate(x.assign_sub(lv)), value)

        # Assign Python floats
        assert_almost_equal(evaluate(x.assign(0.0)), 0.0)
        assert_almost_equal(evaluate(x.assign(latent_value)), value)
        assert_almost_equal(evaluate(x.assign_add(latent_value)), value * 2)
        assert_almost_equal(evaluate(x.assign_sub(latent_value)), value)

        # Use the tf.assign functions instead of the var.assign methods.
        assert_almost_equal(evaluate(tf.compat.v1.assign(x, 0.0)), 0.0)
        assert_almost_equal(evaluate(tf.compat.v1.assign(x, latent_value)), value)
        assert_almost_equal(
            evaluate(tf.compat.v1.assign_add(x, latent_value)), value * 2
        )
        assert_almost_equal(evaluate(tf.compat.v1.assign_sub(x, latent_value)), value)

    # Assign multiple times
    if version.parse(tf.__version__) >= version.parse("2.2") and (
        tf.executing_eagerly() or not distribute_scope
    ):
        assign = x.assign(0.0)
        assert_almost_equal(evaluate(assign), 0.0)
        assert_almost_equal(evaluate(assign.assign(latent_value)), value)
        assert_almost_equal(
            evaluate(x.assign_add(latent_value).assign_add(latent_value)), value * 3
        )
        assert_almost_equal(evaluate(x), value * 3)
        assert_almost_equal(
            evaluate(x.assign_sub(latent_value).assign_sub(latent_value)), value
        )
        assert_almost_equal(evaluate(x), value)

    # Assign with read_value=False
    assert_almost_equal(evaluate(x.assign(0.0)), 0.0)
    assert evaluate(x.assign(latent_value, read_value=False)) is None
    assert_almost_equal(evaluate(x), value)
    assert evaluate(x.assign_add(latent_value, read_value=False)) is None
    assert_almost_equal(evaluate(x), 2 * value)
    assert evaluate(x.assign_sub(latent_value, read_value=False)) is None
    assert_almost_equal(evaluate(x), value)


@pytest.mark.usefixtures("eager_mode", "distribute_scope")
def test_assign_tf_function(quantized):
    x = QuantizedVariable.from_variable(get_var(0.0), quantizer=lambda x: 2 * x)

    @tf.function
    def run_assign():
        return x.assign(1.0).assign_add(3.0).assign_add(3.0).assign_sub(2.0)

    assert_almost_equal(evaluate(run_assign()), 10.0 if quantized else 5.0)


@pytest.mark.usefixtures("eager_and_graph_mode", "distribute_scope")
def test_assign_op():
    x = QuantizedVariable.from_variable(get_var(0.0), quantizer=lambda x: 2 * x)

    @tf.function
    def func():
        assert x.assign(1.0).op is not None
        assert x.assign_add(1.0).op is not None
        assert x.assign_sub(1.0).op is not None

    func()


@pytest.mark.usefixtures("eager_mode", "distribute_scope")
def test_tf_function_control_dependencies(quantized):
    x = QuantizedVariable.from_variable(get_var(0.0), quantizer=lambda x: 2 * x)

    @tf.function
    def func():
        update = x.assign_add(1.0)
        with tf.control_dependencies([update]):
            x.assign_add(1.0)

    func()
    assert_almost_equal(evaluate(x), 4.0 if quantized else 2.0)


@pytest.mark.usefixtures("eager_and_graph_mode")
def test_checkpoint(tmp_path):
    x = QuantizedVariable.from_variable(get_var(0.0), quantizer=lambda x: 2 * x)
    evaluate(x.initializer)
    evaluate(x.assign(123.0))

    checkpoint = tf.train.Checkpoint(x=x)
    save_path = checkpoint.save(tmp_path)
    evaluate(x.assign(234.0))
    checkpoint.restore(save_path).assert_consumed().run_restore_ops()
    assert isinstance(x, QuantizedVariable)
    assert evaluate(x) == 123.0
    with context.quantized_scope(True):
        assert evaluate(x) == 123.0 * 2


@pytest.mark.usefixtures("distribute_scope")
def test_invalid_wrapped_usage():
    with pytest.raises(ValueError, match="`variable` must be of type"):
        QuantizedVariable.from_variable(tf.constant([1.0]))
    with pytest.raises(ValueError, match="`quantizer` must be `callable` or `None`"):
        QuantizedVariable.from_variable(get_var([1.0]), 1)  # type: ignore
    with pytest.raises(ValueError, match="`precision` must be of type `int` or `None`"):
        QuantizedVariable.from_variable(get_var([1.0]), precision=1.0)  # type: ignore


@pytest.mark.usefixtures("eager_and_graph_mode")
def test_repr(snapshot):
    x = get_var(0.0, name="x")

    class Quantizer:
        def __call__(self, x):
            return x

    snapshot.assert_match(
        repr(QuantizedVariable.from_variable(x, quantizer=lambda x: 2 * x))
    )
    snapshot.assert_match(
        repr(QuantizedVariable.from_variable(x, quantizer=Quantizer()))
    )
    snapshot.assert_match(repr(QuantizedVariable.from_variable(x, precision=1)))


@pytest.mark.usefixtures("eager_mode")
@pytest.mark.parametrize("should_quantize", [True, False])
def test_optimizer(should_quantize):
    x = QuantizedVariable.from_variable(get_var(1.0), quantizer=lambda x: -x)
    opt = tf.keras.optimizers.SGD(1.0)

    def loss():
        with context.quantized_scope(should_quantize):
            return x + 1.0

    @tf.function
    def f():
        opt.minimize(loss, var_list=[x])

    f()
    if should_quantize:
        assert evaluate(x) == 2.0
        with context.quantized_scope(should_quantize):
            assert evaluate(x) == -2.0
    else:
        assert evaluate(x) == 0.0


@pytest.mark.skipif(
    version.parse(tf.__version__) < version.parse("2"),
    reason="Requires TensorFlow 2",
)
def test_saved_model(tmp_path):
    model_path = str(tmp_path / "model")
    x = np.random.normal(size=(4, 32))
    model = testing_utils.get_small_bnn_model(x.shape[1], 16, 10)
    weights = model.get_weights()
    model.save(model_path, save_format="tf")
    reloaded_model = tf.keras.models.load_model(model_path)
    reloaded_weights = reloaded_model.get_weights()
    assert_almost_equal(reloaded_model.predict(x), model.predict(x))
    assert len(reloaded_weights) == len(weights)
    for reloaded_weight, weight in zip(reloaded_weights, weights):
        assert_almost_equal(reloaded_weight, weight)



"""Contains QuantizedVariable, a variable that can be quantized in the forward pass."""
from typing import Optional

import tensorflow as tf
from tensorflow.python.distribute.values import DistributedVariable
from tensorflow.python.framework import ops
from tensorflow.python.ops import resource_variable_ops

from larq import context
from larq.quantizers import Quantizer

# pytype: disable=import-error
try:
    from tensorflow.python.distribute.ps_values import AggregatingVariable
    from tensorflow.python.types.core import Tensor as TensorType
except ModuleNotFoundError:
    TensorType = object
    from tensorflow.python.distribute.values import AggregatingVariable
# pytype: enable=import-error


UNSPECIFIED = object()


class QuantizedVariable(tf.Variable, TensorType):
    """A Variable that can be quantized in the forward pass in applicable contexts."""

    def __init__(
        self,
        variable: tf.Variable,
        quantizer: Optional[Quantizer] = None,
        precision: Optional[int] = None,
        op: Optional[tf.Operation] = UNSPECIFIED,
    ):
        """Creates an QuantizedVariable instance.

        # Arguments
            variable: A floating-point resource variable to wrap.
            quantizer: An optional quantizer to transform the floating-point
                variable to a fake quantized variable.
            precision: An optional integer defining the precision of the quantized
                variable. If `None`, `quantizer.precision` is used.
            op: An optional operation of this variable.
        """
        if not resource_variable_ops.is_resource_variable(variable):
            raise ValueError(
                "`variable` must be of type `tf.ResourceVariable`, "
                f"but got `{type(variable)}`."
            )
        if not (quantizer is None or callable(quantizer)):
            raise ValueError(
                "`quantizer` must be `callable` or `None`, "
                f"but got `{type(quantizer)}`."
            )
        if not (precision is None or type(precision) == int):
            raise ValueError(
                "`precision` must be of type `int` or `None`, "
                f"but got `{type(precision)}`."
            )
        self.latent_variable = variable
        self.quantizer = quantizer
        self.precision = precision or getattr(quantizer, "precision", None)
        self._op = op

    @classmethod
    def from_variable(
        cls,
        variable: tf.Variable,
        quantizer: Optional[Quantizer] = None,
        precision: Optional[int] = None,
        op: Optional[tf.Operation] = UNSPECIFIED,
    ):
        """Creates a QuantizedVariable that wraps another variable.

        This typically just returns `QuantizedVariable(variable)`. But, if the variable
        is a DistributedVariable or one of its subclasses, we instead dynamically
        create a class that subclasses from both QuantizedVariable and
        variable.__class__. This is so the returned variable will still pass
        `isinstance(variable, variable.__class__)`, which is required for
        DistributedVariables and its subclasses to work properly.

        # Arguments
            variable: A floating-point resource variable to wrap.
            quantizer: An optional quantizer to transform the floating-point variable to
                a fake quantized variable.
            precision: An optional integer defining the precision of the quantized
                variable. If `None`, `quantizer.precision` is used.
            op: An optional operation of this variable.

        # Returns
            A QuantizedVariable that wraps the variable.
        """
        if not isinstance(variable, (DistributedVariable, AggregatingVariable)):
            return cls(variable, quantizer, precision, op=op)

        class QuantizedDistributedVariable(cls, variable.__class__):
            """A QuantizedVariable that also subclasses from `variable.__class__`.

            `variable.__class__` is either a `DistributedVariable` or an
            `AggregatingVariable`.
            """

            def get(self, *args, **kwargs):
                # For some reason this is needed to make unit `x + x` pass on TF 1.14
                return self._quantize(self.latent_variable.get(*args, **kwargs))

        return QuantizedDistributedVariable(variable, quantizer, precision, op=op)

    def _quantize(self, value):
        if self.quantizer and context.should_quantize():
            return self.quantizer(value)
        return value

    def value(self):
        return self._quantize(self.latent_variable.value())

    def read_value(self):
        return self._quantize(self.latent_variable.read_value())

    def numpy(self):
        return self._quantize(self.latent_variable).numpy()

    def sparse_read(self, *args, **kwargs):
        return self._quantize(self.latent_variable.sparse_read(*args, **kwargs))

    def gather_nd(self, *args, **kwargs):
        return self._quantize(self.latent_variable.gather_nd(*args, **kwargs))

    def __getattr__(self, name):
        return getattr(self.latent_variable, name)

    def _dense_var_to_tensor(self, *args, **kwargs):
        return self._quantize(
            self.latent_variable._dense_var_to_tensor(*args, **kwargs)
        )

    def eval(self, session=None):
        return self._quantize(self.latent_variable).eval(session=session)

    def initialized_value(self):
        return self._quantize(self.latent_variable.initialized_value())

    @property
    def initial_value(self):
        return self._quantize(self.latent_variable.initial_value)

    def _should_act_as_resource_variable(self):
        """Pass resource_variable_ops.is_resource_variable check."""
        pass

    @staticmethod
    def _get_name(obj) -> str:
        try:
            return obj.__name__
        except AttributeError:
            return obj.__class__.__name__

    def __repr__(self) -> str:
        repr_ = (
            f"<{self.__class__.__name__} '{self.name}' "
            f"shape={self.shape} dtype={self.dtype.name}"
        )
        if self.quantizer is not None:
            repr_ += f" quantizer={self._get_name(self.quantizer)}"
        if self.precision is not None:
            repr_ += f" precision={self.precision}"
        if tf.executing_eagerly() and not self._in_graph_mode:
            return f"{repr_} numpy={ops.numpy_text(self.read_value(), is_repr=True)}>"
        return f"{repr_}>"

    # Method delegations: We delegate the following methods to self.latent_variable.
    # Each of these methods simply calls the same method on self.latent_variable. The
    # base Variable raises NotImplementedError for most of these, so we must
    # override them.
    #
    # We do not define the following methods from Variable for the following
    # reasons:
    #   * 'ref': Instead we inherit the definition from Variable.
    #     If we defined and delegated to Variable, the ref of an QuantizedVariable
    #     would be the same as the ref of the underlying variable, which would be
    #     strange as they are different Python objects.

    def set_shape(self, *args, **kwargs):
        return self.latent_variable.set_shape(*args, **kwargs)

    @property
    def trainable(self):
        return self.latent_variable.trainable

    @property
    def synchronization(self):
        return self.latent_variable.synchronization

    @property
    def aggregation(self):
        return self.latent_variable.aggregation

    @property
    def constraint(self):
        return self.latent_variable.constraint

    def _apply_assign_update(
        self, update_fn, value, use_locking=None, name=None, read_value=True
    ):
        if ops.executing_eagerly_outside_functions():
            assign_op = update_fn(value, use_locking, name, False)
            if read_value:
                return QuantizedVariable.from_variable(
                    self.latent_variable, self.quantizer, self.precision, op=assign_op
                )
            return assign_op

        # Fallback to wrapping the returned variable in graph mode if possible
        assign_var = update_fn(value, use_locking, name, read_value)
        if read_value and resource_variable_ops.is_resource_variable(assign_var):
            return QuantizedVariable.from_variable(
                assign_var, self.quantizer, self.precision
            )
        return assign_var

    def _apply_update(self, update_fn, *args, **kwargs):
        update_var = update_fn(*args, **kwargs)
        if ops.executing_eagerly_outside_functions():
            return self

        # Fallback to wrapping the returned variable in graph mode if possible
        if resource_variable_ops.is_resource_variable(update_var):
            return QuantizedVariable.from_variable(
                update_var, self.quantizer, self.precision
            )
        return update_var

    def assign(self, value, use_locking=None, name=None, read_value=True):
        return self._apply_assign_update(
            self.latent_variable.assign, value, use_locking, name, read_value
        )

    def assign_add(self, delta, use_locking=None, name=None, read_value=True):
        return self._apply_assign_update(
            self.latent_variable.assign_add, delta, use_locking, name, read_value
        )

    def assign_sub(self, delta, use_locking=None, name=None, read_value=True):
        return self._apply_assign_update(
            self.latent_variable.assign_sub, delta, use_locking, name, read_value
        )

    def scatter_sub(self, *args, **kwargs):
        return self._apply_update(self.latent_variable.scatter_sub, *args, **kwargs)

    def scatter_add(self, *args, **kwargs):
        return self._apply_update(self.latent_variable.scatter_add, *args, **kwargs)

    def scatter_max(self, *args, **kwargs):
        return self._apply_update(self.latent_variable.scatter_max, *args, **kwargs)

    def scatter_min(self, *args, **kwargs):
        return self._apply_update(self.latent_variable.scatter_min, *args, **kwargs)

    def scatter_mul(self, *args, **kwargs):
        return self._apply_update(self.latent_variable.scatter_mul, *args, **kwargs)

    def scatter_div(self, *args, **kwargs):
        return self._apply_update(self.latent_variable.scatter_div, *args, **kwargs)

    def scatter_update(self, *args, **kwargs):
        return self._apply_update(self.latent_variable.scatter_update, *args, **kwargs)

    def batch_scatter_update(self, *args, **kwargs):
        return self._apply_update(
            self.latent_variable.batch_scatter_update, *args, **kwargs
        )

    def scatter_nd_sub(self, *args, **kwargs):
        return self._apply_update(self.latent_variable.scatter_nd_sub, *args, **kwargs)

    def scatter_nd_add(self, *args, **kwargs):
        return self._apply_update(self.latent_variable.scatter_nd_add, *args, **kwargs)

    def scatter_nd_update(self, *args, **kwargs):
        return self._apply_update(
            self.latent_variable.scatter_nd_update, *args, **kwargs
        )

    def count_up_to(self, *args, **kwargs):
        return self.latent_variable.count_up_to(*args, **kwargs)

    def load(self, *args, **kwargs):
        return self.latent_variable.load(*args, **kwargs)

    @property
    def dtype(self):
        return self.latent_variable.dtype

    @property
    def name(self):
        return self.latent_variable.name

    @property
    def _shared_name(self):
        return self.latent_variable._shared_name

    @property
    def initializer(self):
        return self.latent_variable.initializer

    @property
    def device(self):
        return self.latent_variable.device

    @property
    def op(self):
        if self._op is not UNSPECIFIED:
            return self._op
        return self.latent_variable.op

    @property
    def graph(self):
        return self.latent_variable.graph

    @property
    def shape(self):
        return self.latent_variable.shape

    def get_shape(self):
        return self.latent_variable.get_shape()

    def _gather_saveables_for_checkpoint(self):
        # By delegating this method to the wrapped variable, checkpoints with
        # QuantizedVariables are identical to checkpoints with normal variables.
        # Therefore models checkpointed with QuantizedVariables can be restored on
        # models with normal variables, and vice versa.
        return self.latent_variable._gather_saveables_for_checkpoint()

    def _map_resources(self, *args):
        # By delegating this method to the wrapped variable, SavedModel with
        # QuantizedVariables are identical to SavedModel with normal variables.
        obj_map, resource_map = self.latent_variable._map_resources(*args)
        obj_map[self] = obj_map[self.latent_variable]
        return obj_map, resource_map

    # TODO: Maybe encode the fact the variable is an QuantizedVariable in to_proto().
    def to_proto(self, *args, **kwargs):
        return self.latent_variable.to_proto(*args, **kwargs)

    def from_proto(self, *args, **kwargs):
        return self.latent_variable.from_proto(*args, **kwargs)

    # Delegate the private attributes _handle_name and _initializer_op to
    # self.latent_variable. SavedModel sets these attributes when loading a model. For
    # example, it sets _handle_name here:
    # https://github.com/tensorflow/tensorflow/blob/db26bd574fa95b5bdd53c08463dd19407cc0297e/tensorflow/python/keras/saving/saved_model/load.py#L211
    # We need to expose these attributes on AutoCastVariable as well for
    # SavedModel to work properly.
    # TODO: Find a better way to support SavedModel. Exposing private attributes is
    # hacky and difficult to maintain.
    # For more info see https://github.com/tensorflow/tensorflow/commit/1fcda57f37c2ac854cabf1c3462eb14e39d36c60
    @property
    def _handle_name(self):
        return self.latent_variable._handle_name

    @_handle_name.setter
    def _handle_name(self, handle_name):
        self.latent_variable._handle_name = handle_name

    @property
    def _initializer_op(self):
        return self.latent_variable._initializer_op

    @_initializer_op.setter
    def _initializer_op(self, initializer_op):
        self.latent_variable._initializer_op = initializer_op

    def _as_graph_element(self):
        if self.quantizer and context.should_quantize():
            return self.quantizer(self.latent_variable)
        graph_element = self.latent_variable._as_graph_element()
        if graph_element is None:
            return self._op
        return graph_element


QuantizedVariable._OverloadAllOperators()
tf.register_tensor_conversion_function(
    QuantizedVariable, QuantizedVariable._dense_var_to_tensor
)
try:
    ops.register_dense_tensor_like_type(QuantizedVariable)
except AttributeError:
    pass
